# cython: language_level=3
#
# Copyright (c) 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in
# the root directory of this source tree.
#


import errno


class DBusBaseError(Exception):
    def __init__(self, errno, err_name=None, err_message=None):
      is_base_error = self.__class__.__name__ == 'DBusBaseError'
      custom_msg = (
        "\nThis is DBusBaseError, a base error for DBus (i bet you did not "
        "see that coming) if you need a special error, enhance "
        "pystemd.sysdexc module!."
      ) if is_base_error else ''

      self.errno = errno
      self.err_name = (err_name or '') if is_base_error else ''
      self.err_message = err_message or ''

      super(DBusBaseError, self).__init__(
            '%s[err %s]: %s%s' % (self.err_name,
                              self.errno,
                              self.err_message, custom_msg))


class DBusFailedError(DBusBaseError):
    pass


class DBusNoMemoryError(DBusBaseError):
    pass


class DBusServiceUnknownError(DBusBaseError):
    pass


class DBusNameHasNoOwnerError(DBusBaseError):
    pass


class DBusNoReplyError(DBusBaseError):
    pass


class DBusIOErrorError(DBusBaseError):
    pass


class DBusBadAddressError(DBusBaseError):
    pass


class DBusNotSupportedError(DBusBaseError):
    pass


class DBusLimitsExceededError(DBusBaseError):
    pass


class DBusAccessDeniedError(DBusBaseError):
    pass


class DBusAuthFailedError(DBusBaseError):
    pass


class DBusNoServerError(DBusBaseError):
    pass


class DBusTimeoutError(DBusBaseError):
    pass


class DBusTimedOutError(DBusBaseError):
    pass


class DBusNoNetworkError(DBusBaseError):
    pass


class DBusAddressInUseError(DBusBaseError):
    pass


class DBusDisconnectedError(DBusBaseError):
    pass


class DBusInvalidArgsError(DBusBaseError):
    pass


class DBusFileNotFoundError(DBusBaseError):
    pass


class DBusFileExistsError(DBusBaseError):
    pass


class DBusUnknownMethodError(DBusBaseError):
    pass


class DBusUnknownObjectError(DBusBaseError):
    pass


class DBusUnknownInterfaceError(DBusBaseError):
    pass


class DBusUnknownPropertyError(DBusBaseError):
    pass


class DBusPropertyReadOnlyError(DBusBaseError):
    pass


class DBusUnixProcessIdUnknownError(DBusBaseError):
    pass


class DBusInvalidSignatureError(DBusBaseError):
    pass


class DBusInconsistentMessageError(DBusBaseError):
    pass


class DBusMatchRuleNotFoundError(DBusBaseError):
    pass


class DBusMatchRuleInvalidError(DBusBaseError):
    pass


class DBusInteractiveAuthorizationRequiredError(DBusBaseError):
    pass


class DBusNoSuchUnitError(DBusBaseError):
    pass


class DBusConnectionRefusedError(DBusBaseError):
    pass


class DBusInterruptedError(DBusBaseError, InterruptedError):
    pass


cdef dict DBUS_ERROR_MAP = {
    b"org.freedesktop.DBus.Error.Failed": DBusFailedError,
    b"org.freedesktop.DBus.Error.NoMemory": DBusNoMemoryError,
    b"org.freedesktop.DBus.Error.ServiceUnknown": DBusServiceUnknownError,
    b"org.freedesktop.DBus.Error.NameHasNoOwner": DBusNameHasNoOwnerError,
    b"org.freedesktop.DBus.Error.NoReply": DBusNoReplyError,
    b"org.freedesktop.DBus.Error.IOError": DBusIOErrorError,
    b"org.freedesktop.DBus.Error.BadAddress": DBusBadAddressError,
    b"org.freedesktop.DBus.Error.NotSupported": DBusNotSupportedError,
    b"org.freedesktop.DBus.Error.LimitsExceeded": DBusLimitsExceededError,
    b"org.freedesktop.DBus.Error.AccessDenied": DBusAccessDeniedError,
    b"org.freedesktop.DBus.Error.AuthFailed": DBusAuthFailedError,
    b"org.freedesktop.DBus.Error.NoServer": DBusNoServerError,
    b"org.freedesktop.DBus.Error.Timeout": DBusTimeoutError,
    b'org.freedesktop.DBus.Error.TimedOut': DBusTimedOutError,
    b"org.freedesktop.DBus.Error.NoNetwork": DBusNoNetworkError,
    b"org.freedesktop.DBus.Error.AddressInUse": DBusAddressInUseError,
    b"org.freedesktop.DBus.Error.Disconnected": DBusDisconnectedError,
    b"org.freedesktop.DBus.Error.InvalidArgs": DBusInvalidArgsError,
    b"org.freedesktop.DBus.Error.FileNotFound": DBusFileNotFoundError,
    b"org.freedesktop.DBus.Error.FileExists": DBusFileExistsError,
    b"org.freedesktop.DBus.Error.UnknownMethod": DBusUnknownMethodError,
    b"org.freedesktop.DBus.Error.UnknownObject": DBusUnknownObjectError,
    b"org.freedesktop.DBus.Error.UnknownInterface": DBusUnknownInterfaceError,
    b"org.freedesktop.DBus.Error.UnknownProperty": DBusUnknownPropertyError,
    b"org.freedesktop.DBus.Error.PropertyReadOnly": DBusPropertyReadOnlyError,
    b"org.freedesktop.DBus.Error.UnixProcessIdUnknown": DBusUnixProcessIdUnknownError,
    b"org.freedesktop.DBus.Error.InvalidSignature": DBusInvalidSignatureError,
    b"org.freedesktop.DBus.Error.InconsistentMessage": DBusInconsistentMessageError,
    b"org.freedesktop.DBus.Error.MatchRuleNotFound": DBusMatchRuleNotFoundError,
    b"org.freedesktop.DBus.Error.MatchRuleInvalid": DBusMatchRuleInvalidError,
    b"org.freedesktop.DBus.Error.InteractiveAuthorizationRequired":
        DBusInteractiveAuthorizationRequiredError,
    b"org.freedesktop.systemd1.NoSuchUnit": DBusNoSuchUnitError,
    b"System.Error.EINTR": DBusInterruptedError,
}


cdef dict OS_ERROR_MAP = {
    -errno.ECONNREFUSED: DBusConnectionRefusedError,
    -errno.ENOENT: DBusFileNotFoundError,
}


def DBusError(err_no, err_name=None, err_message=""):
    return DBUS_ERROR_MAP.get(
        err_name, OS_ERROR_MAP.get(err_no, DBusBaseError)
    )(err_no, err_name, err_message)



#!/usr/bin/env python3
#
# Copyright (c) 2019-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in
# the root directory of this source tree.
#

from unittest import TestCase
from unittest.mock import ANY, call, patch

import pystemd.journal
from pystemd.utils import x2char_star


class TestJournalLog(TestCase):
    MESSAGE = "my message"
    PRIORITY = 101

    @patch("pystemd.journal.sendv")
    def test_msg(self, sendv):
        pystemd.journal.log(self.PRIORITY, self.MESSAGE)
        LOG_LINE = 24
        # ^^ that should be the line we called pystemd.journal.log.

        self.assertEqual(sendv.call_count, 1, "sendv should only be called once")

        sendv.call_args.assert_called_with(
            call(
                CODE_CONTEXT=ANY,
                CODE_LINE=LOG_LINE,
                CODE_FILE=__file__,
                MESSAGE=x2char_star(self.MESSAGE),
                PRIORITY=self.PRIORITY,
                CODE_FUNC="test_msg",
            )
        )



# cython: language_level=3
#
# Copyright (c) 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in
# the root directory of this source tree.
#

import os

cimport pystemd.dbusc as dbusc

from pprint import pformat

from pystemd.dbusexc import DBusError
from pystemd.utils import x2char_star

from libc.stdint cimport (
    int16_t,
    int32_t,
    int64_t,
    uint8_t,
    uint16_t,
    uint32_t,
    uint64_t,
)
from libc.stdlib cimport free
from libcpp cimport bool

CONTAINER_TYPES = (
  dbusc.SD_BUS_TYPE_ARRAY,
  dbusc.SD_BUS_TYPE_VARIANT,
  dbusc.SD_BUS_TYPE_STRUCT,
  dbusc.SD_BUS_TYPE_DICT_ENTRY
)

class VariableReturn(object):
  def __init__(self, parent=None, v_type=None):
    self.data = []
    self.parent = parent
    self.v_type = v_type

  def create_child(self, v_type=None):
    self.data.append(
      VariableReturn(
        parent=self,
        v_type=v_type))

    return self.data[-1]

  def append(self, dt):
    self.data.append(dt)

  def dump(self):
    o = [
      e.dump() if isinstance(e, VariableReturn) else e
      for e in self.data
    ]
    if self.v_type == dbusc.SD_BUS_TYPE_STRUCT:
      return tuple(o)
    elif self.v_type == dbusc.SD_BUS_TYPE_VARIANT:
      return o[0]
    elif self.v_type == dbusc.SD_BUS_TYPE_DICT_ENTRY:
      return {o[0]: o[1]}
    elif self.v_type == dbusc.SD_BUS_TYPE_ARRAY and o and isinstance(o[0], dict):
      return dict(e for d in o for e in d.items())
    return o


cdef class DbusMessage:
    cdef dbusc.sd_bus_message *_msg

    cdef public body
    cdef public dict headers

    def __dealloc__(self):
      dbusc.sd_bus_message_unref(self._msg)

    cdef dbusc.sd_bus_message **ref(self):
        return &(self._msg)

    cdef dbusc.sd_bus_message *msg(self):
        return self._msg

    cdef set_bus_message(self, dbusc.sd_bus_message *msg):
        assert self._msg == NULL
        self._msg = msg
        dbusc.sd_bus_message_ref(msg)

    cpdef process_reply(self, bool with_headers):
        """
          Read through a sd_bus_message reply object and return the answer to that call.
          It's loosely based on bus_message_dump in \
          https://github.com/systemd/systemd/blob/master/src/libsystemd/sd-bus/bus-dump.c
          except that will dump to stdout, this will give you a nice python array (you
          are welcome!).

          The nice thing about this method, is that it will enter complex types like
          arrays or struct. so if the answer is in the form of "a(sbi)", it can return
          as an answer [[b'a', True, 0], [b'b', False, 1], ...]
        """
        cdef:
            int r
            uint8_t msg_type
            char peek_type
            const char *peek_contents = NULL
            dbusc.basic_data rettype

        return_e = VariableReturn()

        while True:
            r = dbusc.sd_bus_message_peek_type(self._msg, &peek_type, &peek_contents)

            if r == 0:
              if return_e.parent is None:
                break

              r = dbusc.sd_bus_message_exit_container(self._msg)

              if r < 0:
                raise DBusError(r, None, "Failed to exit container")

              return_e = return_e.parent
              continue

            if peek_type in CONTAINER_TYPES:
              return_e = return_e.create_child(v_type=peek_type)
              r = dbusc.sd_bus_message_enter_container(self._msg, peek_type, peek_contents)

              if r < 0:
                raise DBusError(r, None, "Failed to enter container")

              continue

            r = dbusc.sd_bus_message_read_basic(self._msg, peek_type, &rettype)

            if r < 0:
              raise DBusError(r, None, "Failed to read message")

            return_e.append(cast_data_to(rettype, peek_type))

        u = return_e.dump()

        if len(u) > 1:
          self.body = u
        elif u:
          self.body = u[0]
        else:
          self.body = None

        # get the headers

        if with_headers:
            msg_type = self.get_type()
            if msg_type == dbusc.SD_BUS_MESSAGE_SIGNAL:
                self.headers = {
                    'Type': msg_type,
                    'Typename': 'signal',
                    #'Endian': None,
                    #'Flags': None,
                    #'Version': None,
                    'Priority': self.get_priority(),
                    'Cookie': self.get_cookie(),
                    'ReplyCookie': None,

                    'Sender':self.get_sender(),
                    'Destination': None,
                    'Path': self.get_path(),
                    'Interface': self.get_interface(),
                    'Member': self.get_member(),

                }
            elif msg_type == dbusc.SD_BUS_MESSAGE_METHOD_CALL:
                self.headers = {
                    'Type': msg_type,
                    'Typename': 'method_call',
                    #'Endian': None,
                    #'Flags': None,
                    #'Version': None,
                    'Priority': self.get_priority(),
                    'Cookie': self.get_cookie(),
                    'ReplyCookie': None,

                    'Sender':self.get_sender(),
                    'Destination': self.get_destination(),
                    'Path': self.get_path(),
                    'Interface': self.get_interface(),
                    'Member': self.get_member(),

                }
            elif msg_type == dbusc.SD_BUS_MESSAGE_METHOD_RETURN:
                self.headers = {
                    'Type': msg_type,
                    'Typename': 'method_return',
                    #'Endian': None,
                    #'Flags': None,
                    #'Version': None,
                    'Priority': self.get_priority(),
                    'Cookie': self.get_cookie(),
                    'ReplyCookie': self.get_reply_cookie(),

                    'Sender':self.get_sender(),
                    'Destination': self.get_destination(),
                    'Path': None,
                    'Interface': None,
                    'Member': None,

                }
            else:
                self.headers = {
                    'Type': msg_type,
                    'Typename': 'Unknown',
                    #'Endian': None,
                    #'Flags': None,
                    #'Version': None,
                    'Priority': None,
                    'Cookie': None,
                    'ReplyCookie': None,

                    'Sender': None,
                    'Destination': None,
                    'Path': None,
                    'Interface': None,
                    'Member': None,

                }

    # interfaces to sd_bus_message
    # this will mostly be available to python also

    cpdef bool is_empty(self):
        return dbusc.sd_bus_message_is_empty(self._msg)

    cpdef bool is_signal(self, const char *interface, const char *member):
        return dbusc.sd_bus_message_is_signal(self._msg, interface, member) > 0


    cpdef uint8_t get_type(self):
        cdef uint8_t ret
        dbusc.sd_bus_message_get_type(self._msg, &ret)
        return ret


    cpdef uint64_t get_cookie(self):
        cdef uint64_t ret
        dbusc.sd_bus_message_get_cookie(self._msg, &ret)
        return ret


    cpdef uint64_t get_reply_cookie(self):
        cdef uint64_t ret
        dbusc.sd_bus_message_get_reply_cookie(self._msg, &ret)
        return ret

    cpdef int64_t get_priority(self):
        cdef int64_t ret
        dbusc.sd_bus_message_get_priority(self._msg, &ret)
        return ret

    cpdef const char *get_path(self):
        return dbusc.sd_bus_message_get_path(self._msg)

    cpdef const char *get_interface(self):
        return dbusc.sd_bus_message_get_path(self._msg)

    cpdef const char *get_member(self):
        return dbusc.sd_bus_message_get_member(self._msg)

    cpdef const char *get_destination(self):
        return dbusc.sd_bus_message_get_destination(self._msg)

    cpdef const char *get_sender(self):
        return dbusc.sd_bus_message_get_sender(self._msg)


cdef class DBus:
    cdef dbusc.sd_bus *bus
    cdef bool user_mode
    cdef int interactive

    def __init__(self, user_mode=False, interactive=None):
      self.user_mode = user_mode
      self.interactive = -1 if interactive is None else int(interactive)

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, *errs):
        self.close()

    cdef int open_dbus_bus(self):
      cdef int r
      # if user_mode not set, then this is easy:
      if not self.user_mode:
        return dbusc.sd_bus_open_system(&(self.bus))
      else:
        return dbusc.sd_bus_open_user(&(self.bus))

    def open(self):
        cdef int rets
        rets = self.open_dbus_bus()
        if (rets < 0):
            raise DBusError(rets, None, "Could not open a bus to DBus")
        if self.interactive >= 0:
            self.set_allow_interactive_authorization(self.interactive)

    def close(self):
        dbusc.sd_bus_close(self.bus)
        dbusc.sd_bus_unref(self.bus)

    def process(self):
        cdef:
            int r
            DbusMessage msg = DbusMessage()

        r = dbusc.sd_bus_process(self.bus, msg.ref())
        if r < 0:
            raise DBusError(r, None, "Failed to process bus")

        return msg

    cdef _msg_append(
        self, dbusc.sd_bus_message *msg_call, int arg_type_i, arg_value):

      cdef char * arg_type_c
      cdef char arg_type = <char>arg_type_i

      if arg_type_i > 0:
          inter = chr(arg_type_i).encode()
          arg_type_c = inter

      if arg_type == dbusc.SD_BUS_TYPE_BYTE:
        r = dbusc.sd_bus_message_append(
          msg_call,
          arg_type_c,
          <uint8_t>arg_value)
      elif arg_type == dbusc.SD_BUS_TYPE_BOOLEAN:
        r = dbusc.sd_bus_message_append(
          msg_call,
          arg_type_c,
          <bool>arg_value)
      elif arg_type == dbusc.SD_BUS_TYPE_INT16:
        r = dbusc.sd_bus_message_append(
          msg_call,
          arg_type_c,
          <int16_t>arg_value)
      elif arg_type == dbusc.SD_BUS_TYPE_UINT16:
        r = dbusc.sd_bus_message_append(
          msg_call,
          arg_type_c,
          <uint16_t>arg_value)
      elif arg_type == dbusc.SD_BUS_TYPE_INT32:
        r = dbusc.sd_bus_message_append(
          msg_call,
          arg_type_c,
          <int32_t>arg_value)
      elif arg_type == dbusc.SD_BUS_TYPE_UINT32:
        r = dbusc.sd_bus_message_append(
          msg_call,
          arg_type_c,
          <uint32_t>arg_value)
      elif arg_type == dbusc.SD_BUS_TYPE_INT64:
        r = dbusc.sd_bus_message_append(
          msg_call,
          arg_type_c,
          <int64_t>arg_value)
      elif arg_type == dbusc.SD_BUS_TYPE_UINT64:
        r = dbusc.sd_bus_message_append(
          msg_call,
          arg_type_c,
          <uint64_t>arg_value)
      elif arg_type == dbusc.SD_BUS_TYPE_DOUBLE:
        r = dbusc.sd_bus_message_append(
          msg_call,
          arg_type_c,
          <double>arg_value)
      elif arg_type in (
          dbusc.SD_BUS_TYPE_STRING, dbusc.SD_BUS_TYPE_OBJECT_PATH,
          dbusc.SD_BUS_TYPE_SIGNATURE):
        carg_value = x2char_star(arg_value)
        r = dbusc.sd_bus_message_append(
          msg_call,
          arg_type_c,
          <char*>carg_value)
      elif arg_type == dbusc.SD_BUS_TYPE_UNIX_FD:
        r = dbusc.sd_bus_message_append(
          msg_call,
          arg_type_c,
          <int>arg_value)
      elif arg_type in CONTAINER_TYPES: # open container
        r = dbusc.sd_bus_message_open_container(
          msg_call,
          arg_type,
          <char*>arg_value)
      elif arg_type_i == -1: # close container
        r = dbusc.sd_bus_message_close_container(msg_call)
      else:
        raise DBusError(
          -42, None, "Unknown arg type %s for %s" % (arg_type, arg_value))

    def call_method(
            self,
            const char *destination,
            const char *path,
            const char *interface,
            const char *method,
            args):

        cdef:
            int r
            int arg_type

            dbusc.sd_bus_message *msg_call = NULL
            dbusc.sd_bus_message *msg_reply = NULL
            dbusc.sd_bus_error error = dbusc.sd_bus_error(NULL, NULL, 0)

            DbusMessage msg = DbusMessage()

        r = dbusc.sd_bus_message_new_method_call(
            self.bus,
            &msg_call,
            destination,
            path,
            interface,
            method
        )

        if r < 0:
            raise DBusError(r, None, 'Could not create DBus method')

        for narg, (arg_type, arg_value) in enumerate(args or []):
            try:
                self._msg_append(msg_call, arg_type, arg_value)
            except TypeError as e:
              raise TypeError(
                str(e) + ' for input "{}", element '
                'number {} in input series:\n{}'.format(
                  arg_value, narg, pformat(args)))


        r = dbusc.sd_bus_call(self.bus, msg_call, 0, &error, msg.ref())

        if r < 0:
            raise DBusError(r, error.name, error.message)

        msg.process_reply(False)

        dbusc.sd_bus_message_unref(msg_call)
        dbusc.sd_bus_message_unref(msg_reply)
        dbusc.sd_bus_error_free(&error)

        return msg


    def get_property(self,
            const char *destination,
            const char *path,
            const char *interface,
            const char *property,
            const char *rtype):

        cdef:
          int r
          dbusc.sd_bus_error error = dbusc.sd_bus_error(NULL, NULL, 0)

          DbusMessage msg = DbusMessage()

        r = dbusc.sd_bus_get_property(
          self.bus,
          destination,
          path,
          interface,
          property,
          &error,
          msg.ref(),
          rtype
        )

        if r < 0:
            raise DBusError(r, error.name, error.message)

        msg.process_reply(False)
        return msg.body

    def match_signal(
      self,
      sender,
      path,
      interface,
      member,
      callback,
      userdata=None
    ):
      """
      This method will register a <callback> (a python function) when the signal
      <member> on <interface> is trigger for the specify interface. An example
      on how to use it can be found in
      https://github.com/systemd/pystemd/blob/master/examples/monitor.py
      """

      IF LIBSYSTEMD_VERSION < 237:
          raise DBusError(-1, "sd_bus_match_signal not available in libsystemd v%d" % LIBSYSTEMD_VERSION)
      ELSE:
          cdef int r

          callback.metadata = {
            'userdata': userdata,
          }

          r = dbusc.sd_bus_match_signal(
            self.bus,
            NULL,
            _b2c(x2char_star(sender)),
            _b2c(x2char_star(path)),
            _b2c(x2char_star(interface)),
            _b2c(x2char_star(member)),
            match_signal_callback_handler,
            <void*> callback
          )

          if r < 0:
              raise DBusError(r, "Failed to add signal match")

    # Direct interface to sd_bus_<methods>

    cpdef wait(self, uint64_t timeout):
        cdef int r = dbusc.sd_bus_wait(self.bus, timeout)
        if r < 0:
            raise DBusError(r, "Failed to wait for bus")

    cpdef const char* get_unique_name(self):
        cdef:
            int r
            const char *unique_name

        r = dbusc.sd_bus_get_unique_name(self.bus, &unique_name)

        if r < 0:
            raise DBusError(r, None, "Failed to get unique name")

        return unique_name

    cpdef int get_fd(self):
      return dbusc.sd_bus_get_fd(self.bus)

    cpdef bool get_allow_interactive_authorization(self):
      return dbusc.sd_bus_get_allow_interactive_authorization(self.bus)

    cpdef set_allow_interactive_authorization(self, bool interactive):
        r = dbusc.sd_bus_set_allow_interactive_authorization(self.bus, interactive)
        if r < 0:
            raise DBusError(r, None, "Failed to set interactive authorization")


cdef class DBusMachine(DBus):
  "DBus class that connects to machine"
  cdef char* machine

  def __init__(self, char* machine):
    self.machine = machine

  cdef int open_dbus_bus(self):
    return dbusc.sd_bus_open_system_machine(&(self.bus), self.machine)


cdef class DBusRemote(DBus):
  "DBus class that connects to a remote host"

  cdef char* host

  def __init__(self, char* host):
    self.host = host

  cdef int open_dbus_bus(self):
    return dbusc.sd_bus_open_system_remote(&(self.bus), self.remote)

cdef class DBusAddress(DBus):
  "DBus class that connects to custom address"
  cdef bytes address
  cdef bint peer_to_peer

  def __init__(self, address, peer_to_peer=False):
    self.address = bytes(address)
    self.peer_to_peer = int(peer_to_peer)

  cdef int open_dbus_bus(self):
    r = dbusc.sd_bus_new(&(self.bus))
    if r < 0:
      return r
    r = dbusc.sd_bus_set_address(self.bus, self.address)

    if r < 0:
        return r

    if not self.peer_to_peer:
      r = dbusc.sd_bus_set_bus_client(self.bus, 1)
      if r < 0:
        return r

    r = dbusc.sd_bus_start(self.bus);
    if r < 0:
      return r

    return 0


cdef cast_data_to(dbusc.basic_data basic, cast_type):
  """
  Get the right casting out of a union, the union type and values can be found
  in: https://www.freedesktop.org/software/systemd/man/sd_bus_message_append.html
  or https://dbus.freedesktop.org/doc/dbus-specification.html#type-system (i
  like the first one more than the second one)
  """
  if cast_type == dbusc.SD_BUS_TYPE_BYTE:
    return basic.u8
  elif cast_type == dbusc.SD_BUS_TYPE_BOOLEAN:
    return <bool>basic.i
  elif cast_type == dbusc.SD_BUS_TYPE_INT16:
    return basic.s16
  elif cast_type == dbusc.SD_BUS_TYPE_UINT16:
    return basic.u16
  elif cast_type == dbusc.SD_BUS_TYPE_INT32:
    return basic.s32
  elif cast_type == dbusc.SD_BUS_TYPE_UINT32:
    return basic.u32
  elif cast_type == dbusc.SD_BUS_TYPE_INT64:
    return basic.s64
  elif cast_type == dbusc.SD_BUS_TYPE_UINT64:
    return basic.u64
  elif cast_type == dbusc.SD_BUS_TYPE_DOUBLE:
    return basic.d64
  elif cast_type == dbusc.SD_BUS_TYPE_STRING:
    return basic.string
  elif cast_type == dbusc.SD_BUS_TYPE_OBJECT_PATH:
    return basic.string
  elif cast_type == dbusc.SD_BUS_TYPE_SIGNATURE:
    return basic.string
  elif cast_type == dbusc.SD_BUS_TYPE_UNIX_FD:
    return basic.i

# Parsing argument library

cpdef int find_closure(char* args, char open, char close):
  cdef int counter = 0
  cdef int n
  cdef char arg

  for n, arg in enumerate(args):
    if arg == open:
      counter += 1
    elif arg == close:
      counter -= 1

    if counter == 0:
      break
  return n


cdef dict COMPILE_METHODS

def compile_simple(char *arg):
  cdef char res = arg[0]
  cdef char *off = &res

  def process_simple(v):
    return [(res, v)]

  return off, process_simple

def compile_array(char *args):
  cdef char array_type = args[1]

  off, pr = COMPILE_METHODS.get(array_type, compile_simple)(args[1:])

  def process_array(v):
    cdef list ret = [(dbusc.SD_BUS_TYPE_ARRAY, off[1:])]
    for i in v:
        ret.extend(pr(i))
    ret.append((-1, None))
    return ret

  off = <char*>b'a' + off
  return off, process_array

def compile_struct(char *args):
  cdef int closing

  closing = find_closure(
    args, dbusc.SD_BUS_TYPE_STRUCT_BEGIN, dbusc.SD_BUS_TYPE_STRUCT_END)

  struc_extend = args[1:closing]

  cs = compile_args(struc_extend)

  def process_struct(v):
    return (
      [(dbusc.SD_BUS_TYPE_STRUCT, struc_extend)] +
      apply_args(cs, list(v)) +
      [(-1, None)]
    )
  return b'(' + struc_extend + b')', process_struct


COMPILE_METHODS = {
  dbusc.SD_BUS_TYPE_ARRAY: compile_array,
  dbusc.SD_BUS_TYPE_STRUCT_BEGIN: compile_struct,
}


cpdef list compile_args(char *args):
  cdef:
    int i = 0
    int size_of_args = len(args)
    char val
    list ret = []

  while i < size_of_args:
    val = args[i]
    off, pr = COMPILE_METHODS.get(val, compile_simple)(args[i:])

    ret.append(pr)
    i += len(off)

  return ret

cpdef list apply_args(list args, list values):
  cdef list ret = []
  for cfunc, value in zip(args, values):
    ret.extend(cfunc(value))
  return ret

cpdef list apply_signature(char *signature, list values):
  cdef list args = compile_args(signature)
  cdef list ret = []
  for cfunc, value in zip(args, values):
    ret.extend(cfunc(value))
  return ret


cpdef bytes path_encode(char* prefix,  char* external_id):
  """Python wrapper for sd_bus_path_encode, it produce a encoded version of a
  systemd path:

  example:

  In [1]: path_encode(b'/org/freedesktop/systemd1/unit', b'postfix.service')
  Out[1]: b'/org/freedesktop/systemd1/unit/postfix_2eservice'

  cdocs: https://www.freedesktop.org/software/systemd/man/sd_bus_path_encode.html
  """

  cdef:
    char* ret_path
    int r

  try:
    r = dbusc.sd_bus_path_encode(prefix, external_id, &ret_path)

    if r < 0:
      return b''

    return ret_path
  finally:
    free(ret_path)


cpdef bytes path_decode(char* path,  char* prefix):
  """Python wrapper for sd_bus_path_decode, it produce a decoded version of a
  systemd path:

  example:

  In [1]: path_decode(
  ...    b'/org/freedesktop/systemd1/unit/postfix_2eservice',
  ...    b'/org/freedesktop/systemd1/unit')
  Out[1]: b'postfix.service'

  cdocs: https://www.freedesktop.org/software/systemd/man/sd_bus_path_encode.html
  """

  cdef:
    char* answer
    int r

  try:
    r = dbusc.sd_bus_path_decode(path, prefix, &answer)

    if r <= 0:
      return b''

    return answer
  finally:
    free(answer)


cdef int match_signal_callback_handler(
  dbusc.sd_bus_message *m,
  void *userdata,
  dbusc.sd_bus_error *ret_error
) except -1:
  cdef:
    DbusMessage msg = DbusMessage()
    object mycallback = <object> userdata

  msg.set_bus_message(m)

  # We can't surface any Exception mycallback may throw, because this is
  # happening as a CALLBACK and we have no access to the caller of this method.
  # This will eventually be resolve into whatever exception the caller decide
  # to raise, usually SystemError
  mycallback(msg, error=None, userdata=mycallback.metadata['userdata'])

  return 0


cdef char* _b2c(bytes py_value):
  "Interface to convert byte (as return by x2char_star) to char* respecting Null"
  cdef char * output = NULL

  if py_value is not None:
    output = py_value

  return output



# Copyright (c) 2020-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in
# the root directory of this source tree.
#
from typing import Any, Dict, List, Tuple

class DbusMessage:
    body: Any
    headers: Dict[str, Any]
    def process_reply(self, with_headers: bool): ...
    def is_empty(self) -> bool: ...
    def is_signal(self, interface: bytes, member: bytes) -> bool: ...

class DBus:
    user_mode: bool
    def __init__(self, user_mode: bool = False) -> None: ...
    def __enter__(self) -> "DBus": ...
    def __exit__(self, *errs) -> None: ...
    def open(self) -> None: ...
    def close(self) -> None: ...
    def process(self) -> DbusMessage: ...
    def call_method(
        self,
        destination: bytes,
        path: bytes,
        interface: bytes,
        method: bytes,
        args: Any,
    ) -> Any: ...
    def get_property(
        self,
        destination: bytes,
        path: bytes,
        interface: bytes,
        property: bytes,
        rtype: bytes,
    ) -> Any: ...
    def match_signal(
        self, sender, path, interface, member, callback, userdata=None
    ) -> None: ...
    def wait(self, timeout: int) -> None: ...
    def get_fd(self) -> int: ...
    def get_allow_interactive_authorization(self) -> bool: ...
    def set_allow_interactive_authorization(self, interactive: bool) -> None: ...

class DBusMachine(DBus):
    def __init__(self, machine: bytes) -> None: ...

class DBusRemote(DBus):
    def __init__(self, host: bytes) -> None: ...

class DBusAddress(DBus):
    def __init__(self, address: bytes, peer_to_peer: bool) -> None: ...

def apply_signature(signature: bytes, values: List[Any]) -> list[Tuple[int, Any]]: ...
def path_encode(prefix: bytes, external_id: bytes) -> bytes: ...
def path_decode(path: bytes, prefix: bytes) -> bytes: ...



#
# Copyright (c) 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in
# the root directory of this source tree.
#


from pystemd.base import SDObject
from pystemd.dbuslib import path_encode
from pystemd.utils import x2char_star


class Machine(SDObject):
    def __init__(self, external_id, bus=None, _autoload=False):
        self.external_id = x2char_star(external_id)
        path = path_encode(b"/org/freedesktop/machine1/machine", self.external_id)
        super(Machine, self).__init__(
            destination=b"org.freedesktop.machine1",
            path=path,
            bus=bus,
            _autoload=_autoload,
        )



#
# Copyright (c) 2020-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in
# the root directory of this source tree.
#

from typing import Any, AnyStr, Iterator, List, Optional, Tuple

from pystemd.base import SDInterface, SDObject
from pystemd.dbuslib import DBus, path_encode
from pystemd.utils import x2char_star

class Unit_Unit(SDInterface):
    ActiveEnterTimestamp: int  # t
    ActiveEnterTimestampMonotonic: int  # t
    ActiveExitTimestamp: int  # t
    ActiveExitTimestampMonotonic: int  # t
    ActiveState: bytes  # s
    After: List[bytes]  # as
    AllowIsolate: bool  # b
    AssertResult: bool  # b
    AssertTimestamp: int  # t
    AssertTimestampMonotonic: int  # t
    Asserts: List[Tuple[bytes, bool, bool, bytes, int]]  # a(sbbsi)
    Before: List[bytes]  # as
    BindsTo: List[bytes]  # as
    BoundBy: List[bytes]  # as
    CanClean: List[bytes]  # as
    CanIsolate: bool  # b
    CanReload: bool  # b
    CanStart: bool  # b
    CanStop: bool  # b
    CollectMode: bytes  # s
    ConditionResult: bool  # b
    ConditionTimestamp: int  # t
    ConditionTimestampMonotonic: int  # t
    Conditions: List[Tuple[bytes, bool, bool, bytes, int]]  # a(sbbsi)
    ConflictedBy: List[bytes]  # as
    Conflicts: List[bytes]  # as
    ConsistsOf: List[bytes]  # as
    DefaultDependencies: bool  # b
    Description: bytes  # s
    Documentation: List[bytes]  # as
    DropInPaths: List[bytes]  # as
    FailureAction: bytes  # s
    FailureActionExitStatus: int  # i
    Following: bytes  # s
    FragmentPath: bytes  # s
    Id: bytes  # s
    IgnoreOnIsolate: bool  # b
    InactiveEnterTimestamp: int  # t
    InactiveEnterTimestampMonotonic: int  # t
    InactiveExitTimestamp: int  # t
    InactiveExitTimestampMonotonic: int  # t
    InvocationID: List[bytes]  # ay
    Job: Tuple[int, bytes]  # (uo)
    JobRunningTimeoutUSec: int  # t
    JobTimeoutAction: bytes  # s
    JobTimeoutRebootArgument: bytes  # s
    JobTimeoutUSec: int  # t
    JoinsNamespaceOf: List[bytes]  # as
    LoadError: Tuple[bytes, bytes]  # (ss)
    LoadState: bytes  # s
    Names: List[bytes]  # as
    NeedDaemonReload: bool  # b
    OnFailure: List[bytes]  # as
    OnFailureJobMode: bytes  # s
    PartOf: List[bytes]  # as
    Perpetual: bool  # b
    PropagatesReloadTo: List[bytes]  # as
    RebootArgument: bytes  # s
    Refs: List[bytes]  # as
    RefuseManualStart: bool  # b
    RefuseManualStop: bool  # b
    ReloadPropagatedFrom: List[bytes]  # as
    RequiredBy: List[bytes]  # as
    Requires: List[bytes]  # as
    RequiresMountsFor: List[bytes]  # as
    Requisite: List[bytes]  # as
    RequisiteOf: List[bytes]  # as
    SourcePath: bytes  # s
    StartLimitAction: bytes  # s
    StartLimitBurst: int  # u
    StartLimitIntervalUSec: int  # t
    StateChangeTimestamp: int  # t
    StateChangeTimestampMonotonic: int  # t
    StopWhenUnneeded: bool  # b
    SubState: bytes  # s
    SuccessAction: bytes  # s
    SuccessActionExitStatus: int  # i
    Transient: bool  # b
    TriggeredBy: List[bytes]  # as
    Triggers: List[bytes]  # as
    UnitFilePreset: bytes  # s
    UnitFileState: bytes  # s
    WantedBy: List[bytes]  # as
    Wants: List[bytes]  # as
    def Start(self, mode: bytes) -> bytes: ...  # s # o
    def Stop(self, mode: bytes) -> bytes: ...  # s # o
    def ResetFailed(self) -> bytes: ...  # s # o
    def Restart(self, mode: bytes) -> bytes: ...  # s # o
    def Reload(self, arg0: bytes) -> bytes: ...  # s  # o
    def TryRestart(self, arg0: bytes) -> bytes: ...  # s  # o
    def ReloadOrRestart(self, arg0: bytes) -> bytes: ...  # s  # o
    def ReloadOrTryRestart(self, arg0: bytes) -> bytes: ...  # s  # o
    def EnqueueJob(
        self, arg0: bytes, arg1: bytes  # s  # s
    ) -> Tuple[
        int, bytes, bytes, bytes, bytes, List[Tuple[int, bytes, bytes, bytes, bytes]]
    ]: ...  # (uososa(uosos))
    def Kill(self, arg0: bytes, arg1: int) -> None: ...  # s  # i
    def SetProperties(
        self, arg0: bool, arg1: List[Tuple[bytes, Any]]  # b  # a(sv)
    ) -> None: ...
    def Ref(self) -> None: ...
    def Unref(self) -> None: ...
    def Clean(self, arg0: List[bytes]) -> None: ...  # as

class Unit_Service(SDInterface):
    AllowedCPUs: List[bytes]  # ay
    AllowedMemoryNodes: List[bytes]  # ay
    AmbientCapabilities: int  # t
    AppArmorProfile: Tuple[bool, bytes]  # (bs)
    BindPaths: List[Tuple[bytes, bytes, bool, int]]  # a(ssbt)
    BindReadOnlyPaths: List[Tuple[bytes, bytes, bool, int]]  # a(ssbt)
    BlockIOAccounting: bool  # b
    BlockIODeviceWeight: List[Tuple[bytes, int]]  # a(st)
    BlockIOReadBandwidth: List[Tuple[bytes, int]]  # a(st)
    BlockIOWeight: int  # t
    BlockIOWriteBandwidth: List[Tuple[bytes, int]]  # a(st)
    BusName: bytes  # s
    CPUAccounting: bool  # b
    CPUAffinity: List[bytes]  # ay
    CPUQuotaPerSecUSec: int  # t
    CPUQuotaPeriodUSec: int  # t
    CPUSchedulingPolicy: int  # i
    CPUSchedulingPriority: int  # i
    CPUSchedulingResetOnFork: bool  # b
    CPUShares: int  # t
    CPUUsageNSec: int  # t
    CPUWeight: int  # t
    CacheDirectory: List[bytes]  # as
    CacheDirectoryMode: int  # u
    CapabilityBoundingSet: int  # t
    CleanResult: bytes  # s
    ConfigurationDirectory: List[bytes]  # as
    ConfigurationDirectoryMode: int  # u
    ControlGroup: bytes  # s
    ControlPID: int  # u
    DefaultMemoryLow: int  # t
    DefaultMemoryMin: int  # t
    Delegate: bool  # b
    DelegateControllers: List[bytes]  # as
    DeviceAllow: List[Tuple[bytes, bytes]]  # a(ss)
    DevicePolicy: bytes  # s
    DisableControllers: List[bytes]  # as
    DynamicUser: bool  # b
    EffectiveCPUs: List[bytes]  # ay
    EffectiveMemoryNodes: List[bytes]  # ay
    Environment: List[bytes]  # as
    EnvironmentFiles: List[Tuple[bytes, bool]]  # a(sb)
    ExecCondition: List[
        Tuple[bytes, List[bytes], bool, int, int, int, int, int, int, int]
    ]  # a(sasbttttuii)
    ExecConditionEx: List[
        Tuple[bytes, List[bytes], bool, int, int, int, int, int, int, int]
    ]  # a(sasbttttuii)
    ExecMainCode: int  # i
    ExecMainExitTimestamp: int  # t
    ExecMainExitTimestampMonotonic: int  # t
    ExecMainPID: int  # u
    ExecMainStartTimestamp: int  # t
    ExecMainStartTimestampMonotonic: int  # t
    ExecMainStatus: int  # i
    ExecReload: List[
        Tuple[bytes, List[bytes], bool, int, int, int, int, int, int, int]
    ]  # a(sasbttttuii)
    ExecReloadEx: List[
        Tuple[bytes, List[bytes], bool, int, int, int, int, int, int, int]
    ]  # a(sasbttttuii)
    ExecStart: List[
        Tuple[bytes, List[bytes], bool, int, int, int, int, int, int, int]
    ]  # a(sasbttttuii)
    ExecStartEx: List[
        Tuple[bytes, List[bytes], bool, int, int, int, int, int, int, int]
    ]  # a(sasbttttuii)
    ExecStartPost: List[
        Tuple[bytes, List[bytes], bool, int, int, int, int, int, int, int]
    ]  # a(sasbttttuii)
    ExecStartPostEx: List[
        Tuple[bytes, List[bytes], bool, int, int, int, int, int, int, int]
    ]  # a(sasbttttuii)
    ExecStartPre: List[
        Tuple[bytes, List[bytes], bool, int, int, int, int, int, int, int]
    ]  # a(sasbttttuii)
    ExecStartPreEx: List[
        Tuple[bytes, List[bytes], bool, int, int, int, int, int, int, int]
    ]  # a(sasbttttuii)
    ExecStop: List[
        Tuple[bytes, List[bytes], bool, int, int, int, int, int, int, int]
    ]  # a(sasbttttuii)
    ExecStopEx: List[
        Tuple[bytes, List[bytes], bool, int, int, int, int, int, int, int]
    ]  # a(sasbttttuii)
    ExecStopPost: List[
        Tuple[bytes, List[bytes], bool, int, int, int, int, int, int, int]
    ]  # a(sasbttttuii)
    ExecStopPostEx: List[
        Tuple[bytes, List[bytes], bool, int, int, int, int, int, int, int]
    ]  # a(sasbttttuii)
    FileDescriptorStoreMax: int  # u
    FinalKillSignal: int  # i
    GID: int  # u
    Group: bytes  # s
    GuessMainPID: bool  # b
    IOAccounting: bool  # b
    IODeviceLatencyTargetUSec: List[Tuple[bytes, int]]  # a(st)
    IODeviceWeight: List[Tuple[bytes, int]]  # a(st)
    IOReadBandwidthMax: List[Tuple[bytes, int]]  # a(st)
    IOReadBytes: int  # t
    IOReadIOPSMax: List[Tuple[bytes, int]]  # a(st)
    IOReadOperations: int  # t
    IOSchedulingClass: int  # i
    IOSchedulingPriority: int  # i
    IOWeight: int  # t
    IOWriteBandwidthMax: List[Tuple[bytes, int]]  # a(st)
    IOWriteBytes: int  # t
    IOWriteIOPSMax: List[Tuple[bytes, int]]  # a(st)
    IOWriteOperations: int  # t
    IPAccounting: bool  # b
    IPAddressAllow: List[Tuple[int, List[bytes], int]]  # a(iayu)
    IPAddressDeny: List[Tuple[int, List[bytes], int]]  # a(iayu)
    IPEgressBytes: int  # t
    IPEgressFilterPath: List[bytes]  # as
    IPEgressPackets: int  # t
    IPIngressBytes: int  # t
    IPIngressFilterPath: List[bytes]  # as
    IPIngressPackets: int  # t
    IgnoreSIGPIPE: bool  # b
    InaccessiblePaths: List[bytes]  # as
    KeyringMode: bytes  # s
    KillMode: bytes  # s
    KillSignal: int  # i
    LimitAS: int  # t
    LimitASSoft: int  # t
    LimitCORE: int  # t
    LimitCORESoft: int  # t
    LimitCPU: int  # t
    LimitCPUSoft: int  # t
    LimitDATA: int  # t
    LimitDATASoft: int  # t
    LimitFSIZE: int  # t
    LimitFSIZESoft: int  # t
    LimitLOCKS: int  # t
    LimitLOCKSSoft: int  # t
    LimitMEMLOCK: int  # t
    LimitMEMLOCKSoft: int  # t
    LimitMSGQUEUE: int  # t
    LimitMSGQUEUESoft: int  # t
    LimitNICE: int  # t
    LimitNICESoft: int  # t
    LimitNOFILE: int  # t
    LimitNOFILESoft: int  # t
    LimitNPROC: int  # t
    LimitNPROCSoft: int  # t
    LimitRSS: int  # t
    LimitRSSSoft: int  # t
    LimitRTPRIO: int  # t
    LimitRTPRIOSoft: int  # t
    LimitRTTIME: int  # t
    LimitRTTIMESoft: int  # t
    LimitSIGPENDING: int  # t
    LimitSIGPENDINGSoft: int  # t
    LimitSTACK: int  # t
    LimitSTACKSoft: int  # t
    LockPersonality: bool  # b
    LogExtraFields: List[List[bytes]]  # aay
    LogLevelMax: int  # i
    LogRateLimitBurst: int  # u
    LogRateLimitIntervalUSec: int  # t
    LogsDirectory: List[bytes]  # as
    LogsDirectoryMode: int  # u
    MainPID: int  # u
    MemoryAccounting: bool  # b
    MemoryCurrent: int  # t
    MemoryDenyWriteExecute: bool  # b
    MemoryHigh: int  # t
    MemoryLimit: int  # t
    MemoryLow: int  # t
    MemoryMax: int  # t
    MemoryMin: int  # t
    MemorySwapMax: int  # t
    MountAPIVFS: bool  # b
    MountFlags: int  # t
    NFileDescriptorStore: int  # u
    NRestarts: int  # u
    NUMAMask: List[bytes]  # ay
    NUMAPolicy: int  # i
    NetworkNamespacePath: bytes  # s
    Nice: int  # i
    NoNewPrivileges: bool  # b
    NonBlocking: bool  # b
    NotifyAccess: bytes  # s
    OOMPolicy: bytes  # s
    OOMScoreAdjust: int  # i
    PAMName: bytes  # s
    PIDFile: bytes  # s
    PassEnvironment: List[bytes]  # as
    Personality: bytes  # s
    PrivateDevices: bool  # b
    PrivateMounts: bool  # b
    PrivateNetwork: bool  # b
    PrivateTmp: bool  # b
    PrivateUsers: bool  # b
    ProtectControlGroups: bool  # b
    ProtectHome: bytes  # s
    ProtectHostname: bool  # b
    ProtectKernelLogs: bool  # b
    ProtectKernelModules: bool  # b
    ProtectKernelTunables: bool  # b
    ProtectSystem: bytes  # s
    ReadOnlyPaths: List[bytes]  # as
    ReadWritePaths: List[bytes]  # as
    ReloadResult: bytes  # s
    RemainAfterExit: bool  # b
    RemoveIPC: bool  # b
    Restart: bytes  # s
    RestartForceExitStatus: Tuple[List[int], List[int]]  # (aiai)
    RestartKillSignal: int  # i
    RestartPreventExitStatus: Tuple[List[int], List[int]]  # (aiai)
    RestartUSec: int  # t
    RestrictAddressFamilies: Tuple[bool, List[bytes]]  # (bas)
    RestrictNamespaces: int  # t
    RestrictRealtime: bool  # b
    RestrictSUIDSGID: bool  # b
    Result: bytes  # s
    RootDirectory: bytes  # s
    RootDirectoryStartOnly: bool  # b
    RootImage: bytes  # s
    RuntimeDirectory: List[bytes]  # as
    RuntimeDirectoryMode: int  # u
    RuntimeDirectoryPreserve: bytes  # s
    RuntimeMaxUSec: int  # t
    SELinuxContext: Tuple[bool, bytes]  # (bs)
    SameProcessGroup: bool  # b
    SecureBits: int  # i
    SendSIGHUP: bool  # b
    SendSIGKILL: bool  # b
    Slice: bytes  # s
    SmackProcessLabel: Tuple[bool, bytes]  # (bs)
    StandardError: bytes  # s
    StandardErrorFileDescriptorName: bytes  # s
    StandardInput: bytes  # s
    StandardInputData: List[bytes]  # ay
    StandardInputFileDescriptorName: bytes  # s
    StandardOutput: bytes  # s
    StandardOutputFileDescriptorName: bytes  # s
    StartupBlockIOWeight: int  # t
    StartupCPUShares: int  # t
    StartupCPUWeight: int  # t
    StartupIOWeight: int  # t
    StateDirectory: List[bytes]  # as
    StateDirectoryMode: int  # u
    StatusErrno: int  # i
    StatusText: bytes  # s
    SuccessExitStatus: Tuple[List[int], List[int]]  # (aiai)
    SupplementaryGroups: List[bytes]  # as
    SyslogFacility: int  # i
    SyslogIdentifier: bytes  # s
    SyslogLevel: int  # i
    SyslogLevelPrefix: bool  # b
    SyslogPriority: int  # i
    SystemCallArchitectures: List[bytes]  # as
    SystemCallErrorNumber: int  # i
    SystemCallFilter: Tuple[bool, List[bytes]]  # (bas)
    TTYPath: bytes  # s
    TTYReset: bool  # b
    TTYVHangup: bool  # b
    TTYVTDisallocate: bool  # b
    TasksAccounting: bool  # b
    TasksCurrent: int  # t
    TasksMax: int  # t
    TemporaryFileSystem: List[Tuple[bytes, bytes]]  # a(ss)
    TimeoutAbortUSec: int  # t
    TimeoutCleanUSec: int  # t
    TimeoutStartUSec: int  # t
    TimeoutStopUSec: int  # t
    TimerSlackNSec: int  # t
    Type: bytes  # s
    UID: int  # u
    UMask: int  # u
    USBFunctionDescriptors: bytes  # s
    USBFunctionStrings: bytes  # s
    UnsetEnvironment: List[bytes]  # as
    User: bytes  # s
    UtmpIdentifier: bytes  # s
    UtmpMode: bytes  # s
    WatchdogSignal: int  # i
    WatchdogTimestamp: int  # t
    WatchdogTimestampMonotonic: int  # t
    WatchdogUSec: int  # t
    WorkingDirectory: bytes  # s
    def GetProcesses(self) -> List[Tuple[bytes, int, bytes]]: ...  # a(sus)
    def AttachProcesses(self, arg0: bytes, arg1: List[int]) -> None: ...  # s  # au

class Unit(SDObject):
    def __init__(
        self, external_id: AnyStr, bus: Optional[DBus] = None, _autoload: bool = False
    ): ...
    def __enter__(self) -> Unit: ...
    Unit: Unit_Unit
    Service: Unit_Service



#
# Copyright (c) 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in
# the root directory of this source tree.
#

import shlex
from pathlib import Path
from typing import Any, Iterable, List, Optional, Tuple, TypeVar, Union

T = TypeVar("T")


def x2char_star(what_to_convert: Any, convert_all: bool = False) -> bytes:
    """
    Converts `what_to_convert` to whatever the platform understand as char*.
    For python2, if this is unicode we turn it into a string. If this is
    python3 and what you pass is a `str` we convert it into `bytes`.

    If `convert_all` is passed we will also convert non string types, so `1`
    will be `b'1'` and `True` will be true
    """

    if isinstance(what_to_convert, Path):
        return str(what_to_convert).encode()
    elif isinstance(what_to_convert, bytes):
        return what_to_convert
    elif isinstance(what_to_convert, str):
        return what_to_convert.encode()
    elif convert_all:
        if isinstance(what_to_convert, bool):
            return str(what_to_convert).lower().encode()
        return repr(what_to_convert).encode()
    else:
        return what_to_convert


def str2cmd(
    cmd: Union[str, bytes], cont: bool = False
) -> Tuple[bytes, Tuple[bytes, ...], bool]:
    """
    coverts a string (or bytes) into a cmd list usable by the Exec* family of Unit Signatures, you could do


    pystemd.run(
        cmd=...,
        extra={
            b"ExecReload": [str2cmd("/bin/echo 'there is no reaload'")]
        },
    )

    The cont param is for ignore failure and continue

    """
    cmd = x2char_star(cmd).decode()
    cmdlist = tuple(x2char_star(_) for _ in shlex.split(cmd))

    return (cmdlist[0], cmdlist, cont)


def strlist2cmd(
    strlist: Iterable[Union[str, bytes]], cont: bool = False
) -> Tuple[bytes, Tuple[bytes, ...], bool]:
    """
    coverts a command (an array of strings or bytes) into a cmd list usable by the Exec* family of Unit Signatures, you could do


    pystemd.run(
        cmd=...,
        extra={
            b"ExecReload": [strlist2cmd(["/bin/echo, 'there is no reaload'])]
        },
    )
    The cont param is for ignore failure and continue

    """
    cmdlist = [x2char_star(_) for _ in strlist]
    return (cmdlist[0], tuple(cmdlist), cont)


def x2cmdlist(
    what_to_convert: Any, cont: bool = False
) -> List[Tuple[bytes, Tuple[bytes, ...], bool]]:
    """
    a really overloaded helper to convert most things into a cmd list that can be passed natevly to the Exec* family. it should

    1. for None and [] return []
    2. for a single string like "/bin/foo bar" will return as [(b"/bin/foo", (b"/bin/foo", "bar"), False)]
    3. for a list|set of strings|bytes, it should interpret that as a single command, and it should just return that as [(cmd[0], cmd, False)]
    4. for a list|set of list|set, it should interpret that as multiple command, and it should just return that as [(cmd[0], cmd, False) for each cmd]

    The last argument is what systemd uses to determine if the command can continue if failed.

    """
    if what_to_convert is None:
        return []

    if isinstance(what_to_convert, (str, bytes)):
        return [str2cmd(what_to_convert, cont)]

    if not isinstance(what_to_convert, (list, tuple)):
        return what_to_convert

    # at this point tyhis is either a list or a tuple
    if not len(what_to_convert):
        return []

    # if the first element is a string then we just woll this in a double array
    if isinstance(what_to_convert[0], (str, bytes)):
        return [strlist2cmd(what_to_convert, cont)]

    return [strlist2cmd(_, cont) for _ in what_to_convert]


def unwrap(obj: Optional[T], msg="object was None") -> T:
    if obj is None:
        raise ValueError(msg)
    return obj



#!/usr/bin/env python3
#
# Copyright (c) 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in
# the root directory of this source tree.
#

from unittest import TestCase

from pystemd.dbuslib import (
    compile_args,
    compile_array,
    compile_simple,
    compile_struct,
    find_closure,
)


class TestCompileSimple(TestCase):
    def test_compile(self):
        signature = b"s"

        off, _ = compile_simple(signature)
        self.assertEqual(signature, off)

    def test_result_of_compile(self):
        signature = b"s"
        call_arg = b"arg"

        _, func = compile_simple(signature)

        self.assertEqual(func(call_arg), [(ord(signature), call_arg)])


class TestCompileArray(TestCase):
    def test_compile_array_with_simple_arg(self):
        signature = b"as"
        call_arg = [b"arg_1", b"arg_2"]

        off, func = compile_array(signature)

        self.assertEqual(signature, off)
        self.assertEqual(
            func(call_arg),
            [
                (ord(b"a"), signature[1:]),
                (ord(b"s"), b"arg_1"),
                (ord(b"s"), b"arg_2"),
                (-1, None),
            ],
        )

    def test_compile_double_array(self):
        signature = b"aas"
        call_arg = [[b"arg_1", b"arg_2"], [b"arg_a", b"arg_b"]]

        off, func = compile_array(signature)

        self.assertEqual(signature, off)
        self.assertEqual(
            func(call_arg),
            [
                (ord(b"a"), signature[1:]),
                (ord(b"a"), b"s"),
                (ord(b"s"), b"arg_1"),
                (ord(b"s"), b"arg_2"),
                (-1, None),
                (ord(b"a"), b"s"),
                (ord(b"s"), b"arg_a"),
                (ord(b"s"), b"arg_b"),
                (-1, None),
                (-1, None),
            ],
        )

    def test_compile_array_of_structs(self):
        signature = b"a(ss)"
        call_arg = [(b"arg_1", b"arg_2"), (b"arg_a", b"arg_b")]

        off, func = compile_array(signature)

        self.assertEqual(signature, off)
        self.assertEqual(
            func(call_arg),
            [
                (ord(b"a"), signature[1:]),
                (ord(b"r"), b"ss"),
                (ord(b"s"), b"arg_1"),
                (ord(b"s"), b"arg_2"),
                (-1, None),
                (ord(b"r"), b"ss"),
                (ord(b"s"), b"arg_a"),
                (ord(b"s"), b"arg_b"),
                (-1, None),
                (-1, None),
            ],
        )


class TestCompileStruct(TestCase):
    def test_compile(self):
        signature = b"(ss)"
        off, _ = compile_struct(signature)
        self.assertEqual(signature, off)

    def test_apply(self):
        signature = b"(ss)"
        call_arg = (b"arg_1", b"arg_2")

        _, func = compile_struct(signature)

        self.assertEqual(
            func(call_arg),
            [
                (ord(b"r"), signature[1:-1]),
                (ord(b"s"), b"arg_1"),
                (ord(b"s"), b"arg_2"),
                (-1, None),
            ],
        )


class TestCompileMainFunc(TestCase):
    def test_compile_array(self):
        signature = b"as"
        funcs = compile_args(signature)
        self.assertEqual(len(funcs), 1)

    def test_compile_simplest(self):
        signature = b"ssibb"
        funcs = compile_args(signature)
        self.assertEqual(len(funcs), len(signature))

    def test_compile_little_complex_expr(self):
        signature = b"s(si)abb"
        funcs = compile_args(signature)
        self.assertEqual(len(funcs), 4)
        self.assertEqual(
            [f.__name__ for f in funcs],
            ["process_simple", "process_struct", "process_array", "process_simple"],
        )


class TestFindClosure(TestCase):
    def test_closure(self):
        self.assertEqual(find_closure(b"(ss)", ord("("), ord(")")), 3)

    def test_multi_closure(self):
        self.assertEqual(find_closure(b"(s(s)aa(((s)))u)sss", ord("("), ord(")")), 15)



#!/usr/bin/env python3
#
# Copyright (c) 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in
# the root directory of this source tree.
#

import pathlib
from unittest import TestCase
from unittest.mock import sentinel

from pystemd.utils import x2char_star, x2cmdlist


class TestContextToCharStar(TestCase):
    def test_pass_normal_vars(self):
        for elem in (0, b"hi", True, [], {}, (3, 4), {3, 4}):
            self.assertEqual(elem, x2char_star(elem))

    def test_convert_to_char(self):
        for elem in ("", "hi all"):
            self.assertEqual(elem.encode(), x2char_star(elem))

        self.assertEqual(b"/this/is/path", x2char_star(pathlib.Path("/this/is/path")))

    def test_convert_all(self):
        self.assertEqual(b"true", x2char_star(True, convert_all=True))
        self.assertEqual(b"false", x2char_star(False, convert_all=True))
        self.assertEqual(b"1", x2char_star(1, convert_all=True))
        self.assertEqual(b"3.14159", x2char_star(3.14159, convert_all=True))
        self.assertEqual(b"100", x2char_star(100, convert_all=True))


class TestX2x2Cmdlist(TestCase):
    def tests_none(self):
        self.assertEqual(x2cmdlist(None), [])
        self.assertEqual(x2cmdlist([]), [])

    def tests_strings(self):
        self.assertEqual(
            x2cmdlist("foo", sentinel.i_pass), [(b"foo", (b"foo",), sentinel.i_pass)]
        )
        self.assertEqual(
            x2cmdlist("foo bar", sentinel.i_pass),
            [(b"foo", (b"foo", b"bar"), sentinel.i_pass)],
        )

    def tests_bytes(self):
        self.assertEqual(
            x2cmdlist(b"foo", sentinel.i_pass), [(b"foo", (b"foo",), sentinel.i_pass)]
        )
        self.assertEqual(
            x2cmdlist(b"foo bar", sentinel.i_pass),
            [(b"foo", (b"foo", b"bar"), sentinel.i_pass)],
        )

    def test_array(self):
        self.assertEqual(
            x2cmdlist(
                [
                    [b"foo", "bar"],
                    [b"bing", "bang"],
                ],
                sentinel.i_pass,
            ),
            [
                (b"foo", (b"foo", b"bar"), sentinel.i_pass),
                (b"bing", (b"bing", b"bang"), sentinel.i_pass),
            ],
        )

        self.assertEqual(
            x2cmdlist([b"foo", "bar"], sentinel.i_pass),
            [(b"foo", (b"foo", b"bar"), sentinel.i_pass)],
        )



#
# Copyright (c) 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in
# the root directory of this source tree.
#

from .manager import Manager
from .unit import Unit

__all__ = ["Manager", "Unit"]



# Copyright (c) 2020-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in
# the root directory of this source tree.
#

from typing import Union

LISTEN_FDS_START: int

class PystemdDaemonError(Exception):
    pass

def listen_fds(unset_environment: bool) -> int: ...
def notify(unset_environment: bool, *args: str, **kwargs: Union[int, str]) -> None: ...
def booted() -> bool: ...
def watchdog_enabled(unset_environment: bool) -> int: ...



#
# Copyright (c) 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in
# the root directory of this source tree.
#

from pystemd.base import SDObject, overwrite_interface_method
from pystemd.dbuslib import apply_signature
from pystemd.systemd1.unit_signatures import signature_array


class Manager(SDObject):
    def __init__(self, bus=None, _autoload=False):
        super(Manager, self).__init__(
            destination=b"org.freedesktop.systemd1",
            path=b"/org/freedesktop/systemd1",
            bus=bus,
            _autoload=_autoload,
        )

    @overwrite_interface_method("org.freedesktop.systemd1.Manager")
    def SetUnitProperties(self, interface_name, unit_name, runtime, properties):
        assert interface_name == b"org.freedesktop.systemd1.Manager"
        args = apply_signature(b"sb", [unit_name, runtime])

        args += signature_array(properties)

        with self.bus_context() as bus:
            return bus.call_method(
                self.destination, self.path, interface_name, b"SetUnitProperties", args
            ).body

    @overwrite_interface_method("org.freedesktop.systemd1.Manager")
    def StartTransientUnit(
        self, interface_name, name, smode, properties, extra_units=None
    ):
        assert interface_name == b"org.freedesktop.systemd1.Manager"

        args = apply_signature(b"ss", [name, smode])
        args += signature_array(properties)

        # extra units
        args += [(ord(b"a"), b"(sa(sv))")]
        for eu_name, eu_properties in extra_units or []:
            args += [(ord(b"r"), b"sa(sv)")]
            args += apply_signature(b"s", [eu_name])
            args += signature_array(eu_properties)
            args += [(-1, None)]

        args += [(-1, None)]

        with self.bus_context() as bus:
            return bus.call_method(
                self.destination, self.path, interface_name, b"StartTransientUnit", args
            ).body



#
# Copyright (c) 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in
# the root directory of this source tree.
#

import re
from contextlib import contextmanager

from lxml import etree

from pystemd.dbuslib import DBus, apply_signature
from pystemd.utils import x2char_star


class SDObject(object):
    def __init__(self, destination, path, bus=None, _autoload=False):
        self.destination = x2char_star(destination)
        self.path = x2char_star(path)

        self._interfaces = {}
        self._loaded = False
        self._bus = bus

        if _autoload:
            self.load()

    def __enter__(self):
        self.load()
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        pass

    def __getattr__(self, name):
        """
        This methods allow us to call properties and methods from the interfaces
        directly from the SDObject, this way both are equivalents:

            with SDObject(destination, path) as u:
                assert u.Interface.Property1 == u.Property1

        """

        # At some point we should verify that 2 interface do not have the same
        # methods and/or properties. But in the meantime, we can trust that
        # the fine folks at systemd will not do this. they have actually go
        # out of their way to make this true:
        # https://github.com/systemd/systemd/blob/119f0f2876ea340cc41525e844487aa88551c219/src/core/dbus-unit.c#L1738-L1746

        for interface in self._interfaces.values():
            if name in (interface.properties + interface.methods):
                return getattr(interface, name)
        raise AttributeError()

    @contextmanager
    def bus_context(self):
        close_bus_at_end = self._bus is None
        try:
            if self._bus is None:
                bus = DBus()
                bus.open()
            else:
                bus = self._bus
            yield bus
        finally:
            if close_bus_at_end:
                bus.close()

    def get_introspect_xml(self):
        with self.bus_context() as bus:
            xml_doc = etree.fromstring(
                bus.call_method(
                    self.destination,
                    self.path,
                    b"org.freedesktop.DBus.Introspectable",
                    b"Introspect",
                    [],
                ).body
            )

            return xml_doc

    def load(self, force=False):
        if self._loaded and not force:
            return

        unit_xml = self.get_introspect_xml()
        decoded_destination = self.destination.decode()

        for interface in unit_xml.getchildren():
            if not etree.iselement(interface):
                continue

            if interface.tag != "interface":
                # This is not bad, we just can't act on this info.
                continue

            interface_name = interface.attrib.get("name")

            self._interfaces[interface_name] = meta_interface(interface)(
                self, interface_name
            )

            if interface_name.startswith(decoded_destination):
                setattr(
                    self,
                    interface_name[len(decoded_destination) + 1 :],
                    self._interfaces[interface_name],
                )
            elif interface_name == "org.freedesktop.DBus.Properties":
                self.Properties = self._interfaces[interface_name]


class SDInterface(object):
    def __init__(self, sd_object, interface_name):
        self.sd_object = sd_object
        self.interface_name = x2char_star(interface_name)

    def __repr__(self):
        return "<%s of %s>" % (self.interface_name, self.sd_object.path.decode())

    def _get_property(self, property_name):
        prop_type = self._properties_xml[property_name].attrib.get("type")
        with self.sd_object.bus_context() as bus:
            return bus.get_property(
                self.sd_object.destination,
                self.sd_object.path,
                self.interface_name,
                x2char_star(property_name),
                x2char_star(prop_type),
            )

    def _set_property(self, property_name, value):
        prop_access = self._properties_xml[property_name].attrib.get("access")
        if prop_access == "read":
            raise AttributeError("{} is read-only".format(property_name))
        else:
            raise NotImplementedError("have not implemented set property")

    def _call_method(self, method_name, *args):
        # If the method exist in the sd_object, and it has been authorized to
        # overwrite the method in this interface, call that one
        overwrite_method = getattr(self.sd_object, method_name, None)
        overwrite_interfaces = getattr(overwrite_method, "overwrite_interfaces", [])

        if callable(overwrite_method) and self.interface_name in overwrite_interfaces:
            return overwrite_method(self.interface_name, *args)

        # There is no overwrite in the sd_object, we should call original method
        # we should call the default method (good enough fpor most cases)
        meth = self._methods_xml[method_name]
        in_args = [
            arg.attrib.get("type")
            for arg in meth.getchildren()
            if etree.iselement(arg) and arg.attrib.get("direction") == "in"
        ]

        return self._auto_call_dbus_method(method_name, in_args, *args)

    def _auto_call_dbus_method(self, method_name, in_args, *args):
        if len(args) != len(in_args):
            raise TypeError(
                "method %s require %s arguments, %s supplied"
                % (method_name, len(in_args), len(args))
            )

        block_chars = re.compile(r"v|\{")
        if any(any(block_chars.finditer(arg)) for arg in in_args):
            raise NotImplementedError(
                "still not implemented methods with complex " "arguments"
            )

        in_signature = "".join(in_args)
        call_args = apply_signature(x2char_star(in_signature), list(args))

        with self.sd_object.bus_context() as bus:
            return bus.call_method(
                self.sd_object.destination,
                self.sd_object.path,
                self.interface_name,
                x2char_star(method_name),
                call_args,
            ).body


def _wrap_call_with_name(func, name):
    def _call(self, *args):
        return func(self, name, *args)

    return _call


def extend_class_def(cls, metaclass):
    """extend cls with metaclass"""

    orig_vars = cls.__dict__.copy()
    slots = orig_vars.get("__slots__")
    if slots is not None:
        if isinstance(slots, str):
            slots = [slots]
        for slots_var in slots:
            orig_vars.pop(slots_var)
    orig_vars.pop("__dict__", None)
    orig_vars.pop("__weakref__", None)
    return metaclass(cls.__name__, cls.__bases__, orig_vars)


def meta_interface(interface):
    class _MetaInterface(type):
        def __new__(metacls, classname, baseclasses, attrs):
            attrs.update(
                {
                    "__xml_dom": interface,
                    "properties": [],
                    "methods": [],
                    "_properties_xml": {},
                    "_methods_xml": {},
                }
            )

            _call_method = attrs["_call_method"]
            _get_property = attrs["_get_property"]
            _set_property = attrs["_set_property"]
            elements = [n for n in interface.getchildren() if etree.iselement(n)]

            for element in elements:
                if element.tag == "property":
                    property_name = element.attrib.get("name")

                    attrs["properties"].append(property_name)
                    attrs["_properties_xml"][property_name] = element
                    attrs[property_name] = property(
                        _wrap_call_with_name(_get_property, property_name),
                        _wrap_call_with_name(_set_property, property_name),
                    )

                elif element.tag == "method":
                    method_name = element.attrib.get("name")
                    attrs["methods"].append(method_name)
                    attrs["_methods_xml"][method_name] = element

                    attrs[method_name] = _wrap_call_with_name(_call_method, method_name)

                    attrs[method_name].__name__ = method_name

            return type.__new__(metacls, classname, baseclasses, attrs)

    return extend_class_def(SDInterface, _MetaInterface)


def overwrite_interface_method(interface):
    "This decorator will sign a method to overwrite a method in a interface"

    def overwrite(func):
        overwrite_interfaces = getattr(func, "overwrite_interfaces", [])
        overwrite_interfaces.append(interface.encode())
        func.overwrite_interfaces = overwrite_interfaces
        return func

    return overwrite



#
# Copyright (c) 2021-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in
# the root directory of this source tree.
#

from typing import Any

from pystemd import DBus, __version__, machine1, systemd1

SDUnit = systemd1.Unit
SDManager = systemd1.Manager
SDMachine = machine1.Machine

def run(
    cmd: Any,
    address=...,
    service_type=...,
    name=...,
    user=...,
    user_mode=...,
    nice=...,
    runtime_max_sec=...,
    env=...,
    extra=...,
    cwd=...,
    machine=...,
    wait=False,
    remain_after_exit=False,
    collect=False,
    raise_on_fail=False,
    pty=...,
    pty_master=...,
    pty_path=...,
    stdin=...,
    stdout=...,
    stderr=...,
    _wait_polling=...,
    slice_=...,
    stop_cmd=...,
    stop_post_cmd=...,
    start_pre_cmd=...,
    start_post_cmd=...,
): ...



import os
from concurrent.futures import ProcessPoolExecutor
from multiprocessing import Process
from multiprocessing.context import BaseContext
from pathlib import Path
from typing import Any, Dict, Optional, Sequence, cast

import psutil

import pystemd.cutils
import pystemd.run
import pystemd.utils


class TransientUnitContext(BaseContext):
    """
    Acts as a multiprocessing.context but its systemd unit aware.
    """

    _name = "fork"

    def __init__(
        self,
        properties: Dict[str, Any],
        main_process: Sequence[str] = (),
    ) -> None:
        self.unit: Optional[pystemd.systemd1.Unit] = None
        self.properties = properties
        self.main_process_cmd = main_process or [
            "/bin/bash",
            "-c",
            "exec sleep infinity",
        ]
        super().__init__()

    def start_unit(self) -> pystemd.systemd1.Unit:
        assert self.unit is None, "Unit already started"
        self.unit = pystemd.run(
            self.main_process_cmd,
            user_mode=False,
            extra={
                **self.properties,
                "Delegate": True,
            },
        )
        return cast(pystemd.systemd1.Unit, self.unit)

    def stop_unit(self) -> None:
        pystemd.utils.unwrap(self.unit, "unit not started").Unit.Stop(b"replace")

    def Process(self, **kwargs) -> "ProcessFromTransientUnit":
        return ProcessFromTransientUnit(
            **kwargs, unit=pystemd.utils.unwrap(self.unit, "unit not started")
        )


def enter_unit(unit):
    """
    This method is meant to be called when you want the current code to moved to a systemd "unit"
    usually will be called before calling run in a process, and its meant to bootstrap the process
    after fork, but before code to run. It basically do:

        1) Attach itself to the main transient unit (making it part of the cgroup)
        2) Change its namespace too all the namespaces of the main transient unit, this is best effort
        3) Change group and user to the group and user of the main transient unit
    """
    # get the unit information
    main_pid = unit.Service.MainPID

    p = psutil.Process(main_pid)
    uid = p.uids().real
    gid = p.gids().real

    unit.Service.AttachProcesses("/", [os.getpid()])

    # attach ourselves to the namespace
    for ns in Path(f"/proc/{main_pid}/ns").iterdir():
        pystemd.cutils.setns(os.open(ns, os.O_RDONLY), 0)

    # change the user
    os.setgid(gid)
    os.setuid(uid)


class _ProcessWithPreRun(Process):
    """
    Parent class that its just a subprocess, but it execute a pre_run hook
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.original_run = self.run
        self.run = self.pre_run  # type: ignore

    def pre_run(self):
        self.original_run()


class ProcessFromTransientUnit(_ProcessWithPreRun):
    """
    Given a unit, this class will execute target (or the run method) as if it where part of that unit.
    This class is meant to be use when the systemd unit has already started. Once this process has stop,
    we wont stop the unit.
    """

    def __init__(self, *, unit=None, **kwargs) -> None:
        super().__init__(**kwargs)
        self.unit = unit

    def pre_run(self):
        enter_unit(self.unit)
        super().pre_run()


class TransientUnitProcess(_ProcessWithPreRun):
    """
    Given a set of properties, this class will execute target (or the run method) as if it where part of that unit.
    this class will start a unit on its own, and then move the process inside the unit. Then when the process has finish
    the unit will finish.
    """

    def __init__(self, *, properties=None, **kwargs) -> None:
        super().__init__(**kwargs)
        self.properties = {
            pystemd.utils.x2char_star(k): v for k, v in (properties or {}).items()
        }

    def pre_run(self):
        context = TransientUnitContext(
            properties=self.properties,
            main_process=[
                "/bin/bash",
                "-c",
                f"exec tail --pid={self.pid} -f /dev/null",
            ],
        )
        unit = context.start_unit()
        enter_unit(unit)
        super().pre_run()


class TransientUnitPoolExecutor(ProcessPoolExecutor):
    """
    Same as concurrent.futures.ProcessPoolExecutor but it makes sure the process get
    started in a systemd transient unit
    """

    def __init__(self, properties, **kwargs):
        self.pool_transient_unit_context = TransientUnitContext(properties)
        super().__init__(**{**kwargs, "mp_context": self.pool_transient_unit_context})

    def __enter__(self):
        self.unit = self.pool_transient_unit_context.start_unit()
        return super().__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            return super().__exit__(exc_type, exc_val, exc_tb)
        finally:
            self.pool_transient_unit_context.stop_unit()


def run(fnc, properties, *args, **kwargs):
    """
    Simple helper to call a method in a single worker
    """
    with TransientUnitPoolExecutor(properties=properties or {}, max_workers=1) as poold:
        future = poold.submit(fnc, *args, **kwargs)
    return future.result()



#
# Copyright (c) 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in
# the root directory of this source tree.
#

import sys

# during development this version is always at least "one up" the
# latest release.
__version__ = "0.13.2"

sys.modules[__name__] = __version__  # type: ignore



#!/usr/bin/env fbpython
# (c) Meta Platforms, Inc. and affiliates. Confidential and proprietary.

from __future__ import annotations

import os
import unittest
from unittest.mock import MagicMock, Mock, patch

import pystemd.futures


@patch.object(pystemd.futures, "TransientUnitPoolExecutor", autospec=True)
class TestFuturesRun(unittest.TestCase):
    def test_run(self, TransientUnitPoolExecutor: MagicMock) -> None:
        function = Mock()
        properties = Mock()

        result = pystemd.futures.run(function, properties=properties, kwargs="kwarg")
        poold = TransientUnitPoolExecutor.return_value.__enter__.return_value
        self.assertEqual(result, poold.submit.return_value.result.return_value)
        TransientUnitPoolExecutor.assert_called_once_with(
            properties=properties, max_workers=1
        )
        poold.submit.assert_called_once_with(function, kwargs="kwarg")


@patch.object(pystemd.futures, "TransientUnitContext", autospec=True)
class TestFuturesPool(unittest.TestCase):
    def test_run(self, TransientUnitContext: MagicMock) -> None:
        properties = Mock()
        context = TransientUnitContext.return_value
        with pystemd.futures.TransientUnitPoolExecutor(properties):
            TransientUnitContext.assert_called_once_with(properties)
            context.start_unit.assert_called_once()
        context.stop_unit.assert_called_once()


@patch("os.setuid", autospec=True)
@patch("os.setgid", autospec=True)
@patch("os.open", autospec=True)
@patch("pystemd.cutils.setns", autospec=True)
@patch.object(pystemd.futures.psutil, "Process", autospec=True)
@patch.object(pystemd.futures, "Path", autospec=True)
class TestEnterUnit(unittest.TestCase):
    def test_enter(
        self,
        Path: MagicMock,
        Process: MagicMock,
        setns: MagicMock,
        os_open: MagicMock,
        os_setgid: MagicMock,
        os_setuid: MagicMock,
    ) -> None:
        unit = Mock()
        main_pid = unit.Service.MainPID
        p = Process.return_value
        uid = p.uids.return_value.real
        gid = p.gids.return_value.real

        pystemd.futures.enter_unit(unit)

        Process.assert_called_once_with(main_pid)
        unit.Service.AttachProcesses.assert_called_once_with("/", [os.getpid()])
        Path.assert_called_once_with(f"/proc/{main_pid}/ns")
        os_setgid.assert_called_once_with(gid)
        os_setuid.assert_called_once_with(uid)


@patch.object(pystemd.futures, "enter_unit", autospec=True)
@patch.object(pystemd.futures, "TransientUnitContext", autospec=True)
class TestTransientUnitProcess(unittest.TestCase):
    def test_pre_run(self, TransientUnitContext: MagicMock, enter_unit: Mock):
        properties = {b"foo": b"bar"}
        target = Mock()
        p = pystemd.futures.TransientUnitProcess(properties=properties, target=target)
        # fake call to per_run
        p.pre_run()
        TransientUnitContext.assert_called_once_with(
            properties=properties,
            main_process=[
                "/bin/bash",
                "-c",
                f"exec tail --pid={p.pid} -f /dev/null",
            ],
        )
        enter_unit.assert_called_once_with(
            TransientUnitContext.return_value.start_unit.return_value
        )



# cython: language_level=3
#
# Copyright (c) 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in
# the root directory of this source tree.
#

# This header includes only the symbols from the v237+ releases of systemd:
# https://raw.githubusercontent.com/systemd/systemd/v237/NEWS

cdef extern from "systemd/sd-bus.h":
  int sd_bus_match_signal(
    sd_bus *bus,
    sd_bus_slot **ret,
    const char *sender,
    const char *path,
    const char *interface,
    const char *member,
    sd_bus_message_handler_t callback,
    void *userdata
  )



#
# Copyright (c) 2020-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in
# the root directory of this source tree.
#

from typing import Any, Dict, List, Optional, Tuple

from pystemd.base import SDInterface, SDObject
from pystemd.dbuslib import DBus

UnitFileChange = Tuple[
    bytes,  # Type of the change (one of symlink or unlink)
    bytes,  # File name of the symlink
    bytes,  # Destination of the symlink
]

class Manager_Manager(SDInterface):
    def StartTransientUnit(
        self,
        unit_name: bytes,
        mode: bytes,
        properties: Dict[bytes, Any],
        extra_units: Optional[List[Tuple[bytes, Dict[bytes, Any]]]],
    ): ...
    def StartUnit(self, unit_name: bytes, mode: bytes): ...
    def StopUnit(self, unit_name: bytes, mode: bytes): ...
    def SetUnitProperties(
        self, unit_name: bytes, runtime: bool, properties: Dict[bytes, Any]
    ): ...
    def EnableUnitFiles(
        self, units: List[bytes], runtime: bool, force: bool
    ) -> Tuple[bool, List[UnitFileChange]]: ...
    def DisableUnitFiles(
        self, units: List[bytes], runtime: bool
    ) -> List[UnitFileChange]: ...
    # autogenerated
    def GetUnit(self, arg0: bytes) -> bytes: ...  # s  # o
    def GetUnitByPID(self, arg0: int) -> bytes: ...  # u  # o
    def GetUnitByInvocationID(self, arg0: List[int]) -> bytes: ...  # ay  # o
    def GetUnitByControlGroup(self, arg0: bytes) -> bytes: ...  # s  # o
    def LoadUnit(self, arg0: bytes) -> bytes: ...  # s  # o
    def StartUnitReplace(
        self, arg0: bytes, arg1: bytes, arg2: bytes  # s  # s  # s
    ) -> bytes: ...  # o
    def ReloadUnit(self, arg0: bytes, arg1: bytes) -> bytes: ...  # s  # s  # o
    def RestartUnit(self, arg0: bytes, arg1: bytes) -> bytes: ...  # s  # s  # o
    def TryRestartUnit(self, arg0: bytes, arg1: bytes) -> bytes: ...  # s  # s  # o
    def ReloadOrRestartUnit(self, arg0: bytes, arg1: bytes) -> bytes: ...  # s  # s  # o
    def ReloadOrTryRestartUnit(
        self, arg0: bytes, arg1: bytes  # s  # s
    ) -> bytes: ...  # o
    def EnqueueUnitJob(
        self, arg0: bytes, arg1: bytes, arg2: bytes  # s  # s  # s
    ) -> Tuple[
        int, bytes, bytes, bytes, bytes, List[Tuple[int, bytes, bytes, bytes, bytes]]
    ]: ...  # (uososa(uosos))
    def KillUnit(self, arg0: bytes, arg1: bytes, arg2: int) -> None: ...  # s  # s  # i
    def CleanUnit(self, arg0: bytes, arg1: List[bytes]) -> None: ...  # s  # as
    def ResetFailedUnit(self, arg0: bytes) -> None: ...  # s
    def RefUnit(self, arg0: bytes) -> None: ...  # s
    def UnrefUnit(self, arg0: bytes) -> None: ...  # s
    def GetUnitProcesses(
        self, arg0: bytes  # s
    ) -> List[Tuple[bytes, int, bytes]]: ...  # a(sus)
    def AttachProcessesToUnit(
        self, arg0: bytes, arg1: bytes, arg2: List[int]  # s  # s  # au
    ) -> None: ...
    def AbandonScope(self, arg0: bytes) -> None: ...  # s
    def GetJob(self, arg0: int) -> bytes: ...  # u  # o
    def GetJobAfter(
        self, arg0: int  # u
    ) -> List[Tuple[int, bytes, bytes, bytes, bytes, bytes]]: ...  # a(usssoo)
    def GetJobBefore(
        self, arg0: int  # u
    ) -> List[Tuple[int, bytes, bytes, bytes, bytes, bytes]]: ...  # a(usssoo)
    def CancelJob(self, arg0: int) -> None: ...  # u
    def ClearJobs(self) -> None: ...
    def ResetFailed(self) -> None: ...
    def ListUnits(
        self,
    ) -> List[
        Tuple[bytes, bytes, bytes, bytes, bytes, bytes, bytes, int, bytes, bytes]
    ]: ...  # a(ssssssouso)
    def ListUnitsFiltered(
        self, arg0: List[bytes]  # as
    ) -> List[
        Tuple[bytes, bytes, bytes, bytes, bytes, bytes, bytes, int, bytes, bytes]
    ]: ...  # a(ssssssouso)
    def ListUnitsByPatterns(
        self, arg0: List[bytes], arg1: List[bytes]  # as  # as
    ) -> List[
        Tuple[bytes, bytes, bytes, bytes, bytes, bytes, bytes, int, bytes, bytes]
    ]: ...  # a(ssssssouso)
    def ListUnitsByNames(
        self, arg0: List[bytes]  # as
    ) -> List[
        Tuple[bytes, bytes, bytes, bytes, bytes, bytes, bytes, int, bytes, bytes]
    ]: ...  # a(ssssssouso)
    def ListJobs(
        self,
    ) -> List[Tuple[int, bytes, bytes, bytes, bytes, bytes]]: ...  # a(usssoo)
    def Subscribe(self) -> None: ...
    def Unsubscribe(self) -> None: ...
    def Dump(self) -> bytes: ...  # s
    def DumpByFileDescriptor(self) -> int: ...  # h
    def Reload(self) -> None: ...
    def Reexecute(self) -> None: ...
    def Exit(self) -> None: ...
    def Reboot(self) -> None: ...
    def PowerOff(self) -> None: ...
    def Halt(self) -> None: ...
    def KExec(self) -> None: ...
    def SwitchRoot(self, arg0: bytes, arg1: bytes) -> None: ...  # s  # s
    def SetEnvironment(self, arg0: List[bytes]) -> None: ...  # as
    def UnsetEnvironment(self, arg0: List[bytes]) -> None: ...  # as
    def UnsetAndSetEnvironment(
        self, arg0: List[bytes], arg1: List[bytes]  # as  # as
    ) -> None: ...
    def ListUnitFiles(self) -> List[Tuple[bytes, bytes]]: ...  # a(ss)
    def ListUnitFilesByPatterns(
        self, arg0: List[bytes], arg1: List[bytes]  # as  # as
    ) -> List[Tuple[bytes, bytes]]: ...  # a(ss)
    def GetUnitFileState(self, arg0: bytes) -> bytes: ...  # s  # s
    def ReenableUnitFiles(
        self, arg0: List[bytes], arg1: bool, arg2: bool  # as  # b  # b
    ) -> Tuple[bool, List[Tuple[bytes, bytes, bytes]]]: ...  # (ba(sss))
    def LinkUnitFiles(
        self, arg0: List[bytes], arg1: bool, arg2: bool  # as  # b  # b
    ) -> List[Tuple[bytes, bytes, bytes]]: ...  # a(sss)
    def PresetUnitFiles(
        self, arg0: List[bytes], arg1: bool, arg2: bool  # as  # b  # b
    ) -> Tuple[bool, List[Tuple[bytes, bytes, bytes]]]: ...  # (ba(sss))
    def PresetUnitFilesWithMode(
        self,
        arg0: List[bytes],  # as
        arg1: bytes,  # s
        arg2: bool,  # b
        arg3: bool,  # b
    ) -> Tuple[bool, List[Tuple[bytes, bytes, bytes]]]: ...  # (ba(sss))
    def MaskUnitFiles(
        self, arg0: List[bytes], arg1: bool, arg2: bool  # as  # b  # b
    ) -> List[Tuple[bytes, bytes, bytes]]: ...  # a(sss)
    def UnmaskUnitFiles(
        self, arg0: List[bytes], arg1: bool  # as  # b
    ) -> List[Tuple[bytes, bytes, bytes]]: ...  # a(sss)
    def RevertUnitFiles(
        self, arg0: List[bytes]  # as
    ) -> List[Tuple[bytes, bytes, bytes]]: ...  # a(sss)
    def SetDefaultTarget(
        self, arg0: bytes, arg1: bool  # s  # b
    ) -> List[Tuple[bytes, bytes, bytes]]: ...  # a(sss)
    def GetDefaultTarget(self) -> bytes: ...  # s
    def PresetAllUnitFiles(
        self, arg0: bytes, arg1: bool, arg2: bool  # s  # b  # b
    ) -> List[Tuple[bytes, bytes, bytes]]: ...  # a(sss)
    def AddDependencyUnitFiles(
        self,
        arg0: List[bytes],  # as
        arg1: bytes,  # s
        arg2: bytes,  # s
        arg3: bool,  # b
        arg4: bool,  # b
    ) -> List[Tuple[bytes, bytes, bytes]]: ...  # a(sss)
    def GetUnitFileLinks(
        self, arg0: bytes, arg1: bool  # s  # b
    ) -> List[bytes]: ...  # as
    def SetExitCode(self, arg0: int) -> None: ...  # y
    def LookupDynamicUserByName(self, arg0: bytes) -> int: ...  # s  # u
    def LookupDynamicUserByUID(self, arg0: int) -> bytes: ...  # u  # s
    def GetDynamicUsers(self) -> List[Tuple[int, bytes]]: ...  # a(us)

class Manager(SDObject):
    def __init__(self, bus: Optional[DBus] = None, _autoload: bool = False): ...
    def __enter__(self) -> Manager: ...
    Manager: Manager_Manager



#
# Copyright (c) 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in
# the root directory of this source tree.
#


from pystemd.base import SDObject


class Manager(SDObject):
    def __init__(self, bus=None, _autoload=False):
        super(Manager, self).__init__(
            destination=b"org.freedesktop.machine1",
            path=b"/org/freedesktop/machine1",
            bus=bus,
            _autoload=_autoload,
        )



#
# Copyright (c) 2020-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in
# the root directory of this source tree.
#

import re
from contextlib import contextmanager
from typing import Any, AnyStr, Iterator, Optional
from xml.dom.minidom import Element

from pystemd.dbuslib import DBus, apply_signature
from pystemd.utils import x2char_star

class SDObject(object):
    def __init__(
        self,
        destination: AnyStr,
        path: AnyStr,
        bus=Optional[DBus],
        _autoload: bool = False,
    ) -> None: ...
    def __enter__(self) -> SDObject: ...
    @contextmanager
    def bus_context(self) -> Iterator[DBus]: ...
    def get_introspect_xml(self) -> Element: ...
    def load(self, force: bool = False) -> None: ...

class SDInterface(object):
    def __init__(self, sd_object: SDObject, interface_name: AnyStr) -> None: ...
    def __repr__(self) -> str: ...
    def _get_property(self, property_name: str) -> Any: ...
    def _set_property(self, property_name: AnyStr, value: Any) -> None: ...



#!/usr/bin/env python3
#
# Copyright (c) 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in
# the root directory of this source tree.
#

from unittest import TestCase
from unittest.mock import MagicMock, patch

from lxml import etree

from pystemd.base import SDObject


class TestContextManager(TestCase):
    def test_context(self):
        with patch.object(SDObject, "load") as load:
            with self.assertRaises(ZeroDivisionError), SDObject(b"d", b"p"):
                raise ZeroDivisionError("we should raise this error")
            # Do not use Mock.assert_called_once(), because its not
            # present in python3.5
            self.assertEqual(load.call_count, 1)


class TestLoad(TestCase):
    def setUp(self):
        xml = etree.fromstring(
            """<node>
            <interface name='non.sysd.interface'></interface>
            <interface name='org.freebeer.obj1.I1'>
                <property name='prop1' type='s'></property>
                <method name='meth1'>
                    <arg direction="in" type="s"/>
                </method>
            </interface>
            </node>"""
        )

        self.introspect_path_xml = xml

    def test_set_load(self):
        obj = SDObject(b"org.freebeer.obj1", b"path", bus=MagicMock())
        obj.get_introspect_xml = lambda: self.introspect_path_xml
        obj.load()

        self.assertIn("non.sysd.interface", obj._interfaces)
        self.assertIn("org.freebeer.obj1.I1", obj._interfaces)

        self.assertIn("prop1", obj.I1.properties)
        obj.I1.prop1  # getting a property
        # Do not use Mock.assert_called_once(), because its not
        # present in python3.5
        self.assertEqual(obj._bus.get_property.call_count, 1)

        self.assertIn("meth1", obj.I1.methods)
        obj.I1.meth1(b"arg1")  # just calling a method
        # Do not use Mock.assert_called_once(), because its not
        # present in python3.5
        self.assertEqual(obj._bus.call_method.call_count, 1)

        with self.assertRaises(TypeError):
            obj.I1.meth1()

        with self.assertRaises(TypeError):
            obj.I1.meth1("arg1", "extra args")


class TestGetItem(TestCase):
    def setUp(self):
        super().setUp()

        class FakeInterface:
            def __init__(self):
                self.properties = ["p"]
                self.methods = ["m"]

                self.p = "<fake property>"

            def m(self):
                return "<fake method>"

        self.destination, self.path = b"com.facebook.pystemd", b"/com/facebook/pystemd"
        self.fake_interface = FakeInterface()
        self.sd_obj = SDObject(self.destination, self.path)
        self.sd_obj._interfaces["myinterface"] = self.fake_interface

    def test_find_in_object(self):
        self.assertEqual(self.sd_obj.destination, self.destination)

    def test_find_property_in_interfaces(self):
        self.assertEqual(self.sd_obj.p, self.fake_interface.p)

    def test_find_method_in_interfaces(self):
        self.assertEqual(self.sd_obj.m(), self.fake_interface.m())



#
# Copyright (c) 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in
# the root directory of this source tree.
#

import fcntl
import os
import pty as ptylib
import struct
import sys
import termios
import tty
import uuid
from contextlib import ExitStack
from selectors import EVENT_READ, DefaultSelector

import pystemd
from pystemd.dbuslib import DBus, DBusAddress, DBusMachine
from pystemd.exceptions import PystemdRunError
from pystemd.systemd1 import Manager as SDManager
from pystemd.systemd1 import Unit
from pystemd.utils import x2char_star, x2cmdlist

EXIT_SUBSTATES = (b"exited", b"failed", b"dead")
USER_MODE = os.getuid() != 0


def get_fno(obj):
    """
    Try to get the best fileno of a obj:
        * If the obj is a integer, it return that integer.
        * If the obj has a fileno method, it return that function call.
    """
    if obj is None:
        return None
    elif isinstance(obj, int):
        return obj
    elif hasattr(obj, "fileno") and callable(obj.fileno):
        return obj.fileno()

    raise TypeError("Expected None, int or fileobject with fileno method")


def run(
    cmd,
    address=None,
    service_type=None,
    name=None,
    user=None,
    user_mode=USER_MODE,
    nice=None,
    runtime_max_sec=None,
    env=None,
    extra=None,
    cwd=None,
    machine=None,
    wait=False,
    remain_after_exit=False,
    collect=False,
    raise_on_fail=False,
    pty=None,
    pty_master=None,
    pty_path=None,
    stdin=None,
    stdout=None,
    stderr=None,
    _wait_polling=None,
    slice_=None,
    stop_cmd=None,
    stop_post_cmd=None,
    start_pre_cmd=None,
    start_post_cmd=None,
):
    """
    pystemd.run imitates systemd-run, but with a pythonic feel to it.

    Options:

        cmd: Array with the command to execute (absolute path only)
        stop_cmd: Array with the command to execute on stop (absolute path only)
        stop_post_cmd: Array with the command to execute after stop (absolute path only)
        start_pre_cmd: Array with the command to execute on pre start (absolute path only)
        start_post_cmd: Array with the command to execute on on post start (absolute path only)
        address: A custom dbus socket address
        service_type: Set the unit type, e.g. notify, oneshot. If you dont give a
            value, the unit type will be whatever systemd thinks is the default.
        name: Name of the unit. If not provided, it will be autogenerated.
        user: Username to execute the command, defaults to current user.
        user_mode: Equivalent to running `systemd-run --user`. Defaults to True
            if current user id not root (uid = 0).
        nice: Nice level to run the command.
        runtime_max_sec: Set seconds before sending a sigterm to the process, if
           the service does not die nicely, it will send a sigkill.
        env: A dict with environment variables.
        extra: If you know what you are doing, you can pass extra configuration
            settings to the start_transient_unit method.
        machine: Machine name to execute the command, by default we connect to
            the host's dbus.
        wait: Wait for command completion before returning control, defaults
            to False.
        remain_after_exit: If True, the transient unit will remain after cmd
            has finished, also if true, this methods will return
            pystemd.systemd1.Unit object. defaults to False and this method
            returns None and the unit will be gone as soon as is done.
        collect: Unload unit after it ran, even when failed.
        raise_on_fail: Will raise a PystemdRunError is cmd exit with non 0
            status code, it won't take affect unless you set wait=True,
            defaults to False.
        pty: Set this variable to True if you want a pty to be created. if you
            pass a `machine`, the pty will be created in the machine. Setting
            this value will ignore whatever you set in pty_master and pty_path.
        pty_master: It has only meaning if you pass a pty_path also, this file
            descriptor will be used to forward redirection to `stdin` and `stdout`
            if no `stdin` or `stdout` is present, then this value does nothing.
        pty_path: Setting this value will pass this pty_path to the created
            process and will connect the process stdin, stdout and stderr to this
            pty. by itself it only ensure that your process has a real pty that
            can have ioctl operation over it. if you also pass a `pty_master`,
            `stdin` and `stdout` the pty forwars is handle for you.
        stdin: Specify a file descriptor for stdin. By default this is `None`
            and your unit will not have a stdin. If you set pty = True, or set a
            `pty_master` then that pty will be read and forwarded to this file
            descriptor.
        stdout: Specify a file descriptor for stdout. By default this is `None`
            and your unit will not have a stdout. If you set pty = True, or set a
            `pty_master` then that pty will be read and forwarded to this file
            descriptor.
        stderr: Specify a file descriptor for stderr. By default this is `None`
            and your unit will not have a stderr.
        slice_: the slice under you want to run the unit.

    More info and examples in:
    https://github.com/systemd/pystemd/blob/master/_docs/pystemd.run.md

    """

    def bus_factory():
        if address:
            return DBusAddress(x2char_star(address))
        elif machine:
            return DBusMachine(x2char_star(machine))
        else:
            return DBus(user_mode=user_mode)

    name = x2char_star(name or "pystemd{}.service".format(uuid.uuid4().hex))
    runtime_max_usec = (runtime_max_sec or 0) * 10**6 or runtime_max_sec

    stdin, stdout, stderr = get_fno(stdin), get_fno(stdout), get_fno(stderr)
    env = env or {}
    unit_properties = {}

    extra = extra or {}
    start_cmd = x2cmdlist(cmd, False) + extra.pop(b"ExecStart", [])
    stop_cmd = x2cmdlist(stop_cmd, False) + extra.pop(b"ExecStop", [])
    stop_post_cmd = x2cmdlist(stop_post_cmd, False) + extra.pop(b"ExecStopPost", [])
    start_pre_cmd = x2cmdlist(start_pre_cmd, False) + extra.pop(b"ExecStartPre", [])
    start_post_cmd = x2cmdlist(start_post_cmd, False) + extra.pop(b"ExecStartPost", [])

    if user_mode:
        _wait_polling = _wait_polling or 0.5

    with ExitStack() as ctexit, DefaultSelector() as sel, bus_factory() as bus, SDManager(
        bus=bus
    ) as manager:
        if pty:
            if machine:
                with pystemd.machine1.Machine(machine) as m:
                    pty_master, pty_path = m.Machine.OpenPTY()
            else:
                pty_master, pty_follower = ptylib.openpty()
                pty_path = os.ttyname(pty_follower).encode()
                ctexit.callback(os.close, pty_master)

        if slice_:
            unit_properties[b"Slice"] = x2char_star(slice_)

        if pty_path:
            unit_properties.update(
                {
                    b"StandardInput": b"tty",
                    b"StandardOutput": b"tty",
                    b"StandardError": b"tty",
                    b"TTYPath": pty_path,
                }
            )

            if None not in (stdin, pty_master):
                # lets set raw mode for stdin so we can forward input without
                # waiting for a new line, but lets also make sure we return the
                # attributes as they where after this method is done
                stdin_attrs = tty.tcgetattr(stdin)
                tty.setraw(stdin)
                ctexit.callback(tty.tcsetattr, stdin, tty.TCSAFLUSH, stdin_attrs)
                sel.register(stdin, EVENT_READ)

            if None not in (stdout, pty_master):
                if os.getenv("TERM"):
                    env[b"TERM"] = env.get(b"TERM", os.getenv("TERM").encode())

                sel.register(pty_master, EVENT_READ)
                # lets be a friend and set the size of the pty.
                winsize = fcntl.ioctl(
                    stdout, termios.TIOCGWINSZ, struct.pack("HHHH", 0, 0, 0, 0)
                )
                fcntl.ioctl(pty_master, termios.TIOCSWINSZ, winsize)
        else:
            unit_properties.update(
                {
                    b"StandardInputFileDescriptor": get_fno(stdin) if stdin else stdin,
                    b"StandardOutputFileDescriptor": get_fno(stdout)
                    if stdout
                    else stdout,
                    b"StandardErrorFileDescriptor": get_fno(stderr)
                    if stderr
                    else stderr,
                }
            )

        unit_properties.update(
            {
                b"Type": service_type,
                b"Description": b"pystemd: " + name,
                b"ExecStartPre": start_pre_cmd or None,
                b"ExecStart": start_cmd,
                b"ExecStartPost": start_post_cmd or None,
                b"ExecStop": stop_cmd or None,
                b"ExecStopPost": stop_post_cmd or None,
                b"RemainAfterExit": remain_after_exit,
                b"CollectMode": b"inactive-or-failed" if collect else None,
                b"WorkingDirectory": cwd,
                b"User": user,
                b"Nice": nice,
                b"RuntimeMaxUSec": runtime_max_usec,
                b"Environment": [
                    b"%s=%s" % (x2char_star(key), x2char_star(value))
                    for key, value in env.items()
                ]
                or None,
            }
        )

        unit_properties.update(extra)
        unit_properties = {k: v for k, v in unit_properties.items() if v is not None}

        unit = Unit(name, bus=bus, _autoload=True)
        if wait:
            mstr = (
                (
                    "type='signal',"
                    "sender='org.freedesktop.systemd1',"
                    "path='{}',"
                    "interface='org.freedesktop.DBus.Properties',"
                    "member='PropertiesChanged'"
                )
                .format(unit.path.decode())
                .encode()
            )

            monbus = bus_factory()
            monbus.open()
            ctexit.callback(monbus.close)

            monitor = pystemd.DBus.Manager(bus=monbus, _autoload=True)
            monitor.Monitoring.BecomeMonitor([mstr], 0)

            monitor_fd = monbus.get_fd()
            sel.register(monitor_fd, EVENT_READ)

        # start the process
        unit_start_job = manager.Manager.StartTransientUnit(
            name, b"fail", unit_properties
        )

        while wait:
            events = sel.select(timeout=_wait_polling)
            _in = [key.fileobj for key, _ in events]

            if stdin in _in:
                data = os.read(stdin, 1024)
                os.write(pty_master, data)

            if pty_master in _in:
                try:
                    data = os.read(pty_master, 1024)
                except OSError:
                    sel.unregister(pty_master)
                else:
                    os.write(stdout, data)

            if monitor_fd in _in:
                m = monbus.process()
                if m.is_empty():
                    continue

                m.process_reply(False)
                if (
                    m.get_path() == unit.path
                    and m.body[0] == b"org.freedesktop.systemd1.Unit"
                ):
                    _, message_job_path = m.body[1].get(b"Job", (0, b"/"))

                    if (
                        message_job_path != unit_start_job
                        and m.body[1].get(b"SubState") in EXIT_SUBSTATES
                    ):
                        break

            if _wait_polling and not _in and unit.Service.MainPID == 0:
                # on usermode the subscribe to events does not work that well
                # this is a temporary hack. you can always not wait on usermode.
                break

        if raise_on_fail:
            if unit.Service.ExecMainStatus:
                raise PystemdRunError(
                    "cmd {} exited with status {}".format(
                        cmd, unit.Service.ExecMainStatus
                    )
                )

        unit.load()
        unit.bus_context = bus_factory
        return unit


# do pystemd.run callable.
run.__module__ = sys.modules[__name__]  # type: ignore
sys.modules[__name__] = run  # type: ignore



# cython: language_level=3
#
# Copyright (c) 2019-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in
# the root directory of this source tree.
#

cimport pystemd.dbusc as dbusc

import inspect

from pystemd.utils import x2char_star

from libc.stdlib cimport free, malloc


def sendv(*args, **kwargs):
  """
  send lines directly to the journal.

example:

```python
import logging
import pystemd.journal

pystemd.journal.sendv(
  f"PRIORITY={logging.INFO}",
  MESSAGE="everything is awesome",
  SYSLOG_IDENTIFIER="tegan"
)
```

  """
  cdef:
    char * text
    int le
    bytes i_str
    dbusc.iovec * iov

  all_options = [
    x2char_star(i)
    for i in args
  ] + [
    b"=".join([x2char_star(key), x2char_star(val, convert_all=True)])
    for key, val in kwargs.items()
  ]

  le = len(all_options)
  iov = <dbusc.iovec *> malloc(le * sizeof(dbusc.iovec))


  for i, iv in enumerate(all_options):
    i_str = iv
    text = i_str
    iov[i].iov_base = text
    iov[i].iov_len = len(text)

  dbusc.sd_journal_sendv(iov, len(all_options))

  free(iov)

def log(int priority, message, **kwargs):
  """
  Send a `message` to the journal with the `priority` specified
  examples:

  ```python
import logging
import pystemd.journal

pystemd.journal.log(
  logging.INFO,
  "everything is awesome",
  SYSLOG_IDENTIFIER="tegan"
)
  ```
  """

  stack = inspect.stack()[0]


  CODE_FILE = stack[1]
  CODE_LINE = stack[2]
  CODE_FUNC = stack[3]
  CODE_CONTEXT = stack[4]

  sendv(
    MESSAGE=x2char_star(message),
    PRIORITY=priority,
    CODE_FILE=CODE_FILE,
    CODE_LINE=CODE_LINE,
    CODE_FUNC=CODE_FUNC,
    CODE_CONTEXT=CODE_CONTEXT,
    **kwargs
  )



#
# Copyright (c) 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in
# the root directory of this source tree.
#


from pystemd.dbuslib import apply_signature
from pystemd.utils import x2char_star

# This is where we add the units and signatures, we steal most of this from
# https://github.com/systemd/systemd/tree/master/src/core/ any of the dbus-*.c
# has that data.

# This dict should have a key, value, where value should be either a byte (
# the signatute) or a callable that returns a tuple of key, signature, value.
# please note that this is not recursive.

KNOWN_UNIT_SIGNATURES = {
    b"Service": b"s",
    b"Requires": b"as",
    b"Requisite": b"as",
    b"Wants": b"as",
    b"BindsTo": b"as",
    b"PartOf": b"as",
    b"RequiredBy": b"as",
    b"RequisiteOf": b"as",
    b"WantedBy": b"as",
    b"BoundBy": b"as",
    b"ConsistsOf": b"as",
    b"Conflicts": b"as",
    b"ConflictedBy": b"as",
    b"Before": b"as",
    b"After": b"as",
    b"OnFailure": b"as",
    b"Triggers": b"as",
    b"TriggeredBy": b"as",
    b"PropagatesReloadTo": b"as",
    b"ReloadPropagatedFrom": b"as",
    b"JoinsNamespaceOf": b"as",
    b"RequiresMountsFor": b"as",
    b"Documentation": b"as",
    b"SourcePath": b"s",
    b"StopWhenUnneeded": b"b",
    b"RefuseManualStart": b"b",
    b"RefuseManualStop": b"b",
    b"AllowIsolate": b"b",
    b"DefaultDependencies": b"b",
    b"OnFailureJobMode": b"s",
    b"IgnoreOnIsolate": b"b",
    b"JobTimeoutAction": b"s",
    b"JobTimeoutRebootArgument": b"s",
    b"Conditions": b"a(sbbsi)",
    b"Asserts": b"a(sbbsi)",
    b"FailureAction": b"s",
    b"SuccessAction": b"s",
    b"RebootArgument": b"s",
    b"CollectMode": b"s",
    b"User": b"s",
    b"Type": b"s",
    b"Group": b"s",
    b"Nice": b"i",
    b"DynamicUser": b"b",
    b"Personality": b"s",
    b"Description": b"s",
    b"NotifyAccess": b"s",
    b"BusName": b"s",
    b"RemainAfterExit": b"b",
    b"NoNewPrivileges": b"b",
    b"RootDirectoryStartOnly": b"b",
    b"PermissionsStartOnly": b"b",
    # exec_command
    b"ExecStartPre": b"a(sasb)",
    b"ExecStart": b"a(sasb)",
    b"ExecStartPost": b"a(sasb)",
    b"ExecReload": b"a(sasb)",
    b"ExecStop": b"a(sasb)",
    b"ExecStopPost": b"a(sasb)",
    # execute properties
    b"UtmpIdentifier": b"s",
    b"UtmpMode": b"s",
    b"PAMName": b"s",
    b"SELinuxContext": b"s",
    b"KeyringMode": b"s",
    b"SyslogLevelPrefix": b"b",
    b"MemoryDenyWriteExecute": b"b",
    b"RestrictRealtime": b"b",
    b"RemoveIPC": b"b",
    b"MountAPIVFS": b"b",
    b"CPUSchedulingResetOnFork": b"b",
    b"LockPersonality": b"b",
    b"SupplementaryGroups": b"as",
    b"SystemCallArchitectures": b"as",
    b"SystemCallFilter": b"(bas)",
    # timeouts
    b"RuntimeMaxUSec": b"t",
    b"RuntimeMaxSec": lambda _, value: (b"RuntimeMaxUSec", b"t", int(value * 10**6)),
    b"WatchdogUSec": b"t",
    b"WatchdogSec": lambda _, value: (b"WatchdogUSec", b"t", int(value * 10**6)),
    b"TimeoutStopUSec": b"t",
    b"TimeoutStopSec": lambda _, value: (
        b"TimeoutStopUSec",
        b"t",
        int(value * 10**6),
    ),
    b"TimeoutStartUSec": b"t",
    b"TimeoutStartSec": lambda _, value: (
        b"TimeoutStartUSec",
        b"t",
        int(value * 10**6),
    ),
    # syslog
    b"SyslogIdentifier": b"s",
    # stdio signatures
    b"StandardInput": b"s",
    b"StandardOutput": b"s",
    b"StandardError": b"s",
    b"TTYPath": b"s",
    b"TTYReset": b"b",
    b"TTYVHangup": b"b",
    b"TTYVTDisallocate": b"b",
    b"IgnoreSIGPIPE": b"b",
    b"StandardInputFileDescriptor": b"h",
    b"StandardOutputFile": b"s",
    b"StandardOutputFileToAppend": b"s",
    b"StandardOutputFileDescriptor": b"h",
    b"StandardErrorFile": b"s",
    b"StandardErrorFileToAppend": b"s",
    b"StandardErrorFileDescriptor": b"h",
    b"StandardInputData": b"ay",
    b"Environment": b"as",
    b"PassEnvironment": b"as",
    b"UnsetEnvironment": b"as",
    b"EnvironmentFiles": b"a(sb)",
    # timer signatures
    b"OnActiveSec": b"t",
    b"RemainAfterElapse": b"b",
    b"OnUnitActiveSec": b"t",
    b"OnCalendar": b"s",
    b"OnStartupSec": b"t",
    b"OnBootSec": b"t",
    b"OnUnitInactiveSec": b"t",
    # paths config
    b"WorkingDirectory": b"s",
    b"RootDirectory": b"s",
    b"RootImage": b"s",
    # binds and paths
    b"BindPaths": b"a(ssbt)",
    b"BindReadOnlyPaths": b"a(ssbt)",
    b"ReadWritePaths": b"as",
    b"ReadOnlyPaths": b"as",
    b"ReadWriteDirectories": b"as",
    b"ReadOnlyDirectories": b"as",
    b"InaccessibleDirectories": b"as",
    b"InaccessiblePaths": b"as",
    b"TemporaryFileSystem": b"a(ss)",
    b"MountFlags": b"t",
    b"StateDirectory": b"as",
    b"CacheDirectory": b"as",
    b"LogsDirectory": b"as",
    b"RuntimeDirectory": b"as",
    b"RuntimeDirectoryPreserve": b"s",
    b"ConfigurationDirectory": b"as",
    b"PrivateDevices": b"b",
    b"PrivateIPC": b"b",
    b"PrivateMounts": b"b",
    b"PrivateNetwork": b"b",
    b"PrivateTmp": b"b",
    b"PrivateUsers": b"b",
    b"ProtectKernelTunables": b"b",
    b"ProtectKernelModules": b"b",
    b"ProtectControlGroups": b"b",
    b"ProtectHome": b"s",
    b"ProtectSystem": b"s",
    # systemd.kill
    b"KillMode": b"s",
    b"KillSignal": b"i",
    b"SendSIGHUP": b"b",
    b"SendSIGKILL": b"b",
    b"RestartPreventExitStatus": b"(aiai)",
    b"RestartForceExitStatus": b"(aiai)",
    b"SuccessExitStatus": b"(aiai)",
    b"Restart": b"s",
    b"RestartUSec": b"t",
    b"RestartSec": lambda _, value: (b"RestartUSec", b"t", int(value * 10**6)),
    # Limits
    b"LimitCPU": b"t",
    b"LimitCPUSoft": b"t",
    b"LimitFSIZE": b"t",
    b"LimitFSIZESoft": b"t",
    b"LimitDATA": b"t",
    b"LimitDATASoft": b"t",
    b"LimitSTACK": b"t",
    b"LimitSTACKSoft": b"t",
    b"LimitCORE": b"t",
    b"LimitCORESoft": b"t",
    b"LimitRSS": b"t",
    b"LimitRSSSoft": b"t",
    b"LimitNOFILE": b"t",
    b"LimitNOFILESoft": b"t",
    b"LimitAS": b"t",
    b"LimitASSoft": b"t",
    b"LimitNPROC": b"t",
    b"LimitNPROCSoft": b"t",
    b"LimitMEMLOCK": b"t",
    b"LimitMEMLOCKSoft": b"t",
    b"LimitLOCKS": b"t",
    b"LimitLOCKSSoft": b"t",
    b"LimitSIGPENDING": b"t",
    b"LimitSIGPENDINGSoft": b"t",
    b"LimitMSGQUEUE": b"t",
    b"LimitMSGQUEUESoft": b"t",
    b"LimitNICE": b"t",
    b"LimitNICESoft": b"t",
    b"LimitRTPRIO": b"t",
    b"LimitRTPRIOSoft": b"t",
    b"LimitRTTIME": b"t",
    b"LimitRTTIMESoft": b"t",
    # cgroup
    b"DevicePolicy": b"s",
    b"Slice": b"s",
    b"Delegate": b"b",
    b"CPUAccounting": b"b",
    b"MemoryAccounting": b"b",
    b"MemoryLow": b"t",
    b"MemoryLowScale": b"u",
    b"MemoryHigh": b"t",
    b"MemoryHighScale": b"u",
    b"MemoryMax": b"t",
    b"MemoryMaxScale": b"u",
    b"MemorySwapMax": b"t",
    b"MemorySwapMaxScale": b"u",
    b"MemoryLimit": b"t",
    b"MemoryLimitScale": b"u",
    b"IOAccounting": b"b",
    b"BlockIOAccounting": b"b",
    b"TasksAccounting": b"b",
    b"TasksMax": b"t",
    b"TasksMaxScale": b"u",
    b"CPUQuota": lambda _, value: (b"CPUQuotaPerSecUSec", b"t", int(value * 10**6)),
    b"CPUQuotaPerSecUSec": b"t",
    b"IPAccounting": b"b",
    b"IPAddressAllow": b"a(iayu)",
    b"IPAddressDeny": b"a(iayu)",
    # Timer properties
    b"TimersMonotonic": b"a(st)",
    b"TimersCalendar": b"a(ss)",
    b"WakeSystem": b"b",
    b"Persistent": b"b",
    # Socket Signatures
    b"BindIPv6Only": b"s",
    b"Backlog": b"u",
    b"TimeoutUSec": b"t",
    b"BindToDevice": b"s",
    b"SocketUser": b"s",
    b"SocketGroup": b"s",
    b"SocketMode": b"u",
    b"DirectoryMode": b"u",
    b"Accept": b"b",
    b"Writable": b"b",
    b"KeepAlive": b"b",
    b"KeepAliveTimeUSec": b"t",
    b"KeepAliveIntervalUSec": b"t",
    b"KeepAliveProbes": b"u",
    b"DeferAcceptUSec": b"t",
    b"NoDelay": b"b",
    b"Priority": b"i",
    b"ReceiveBuffer": b"t",
    b"SendBuffer": b"t",
    b"IPTOS": b"i",
    b"IPTTL": b"i",
    b"PipeSize": b"t",
    b"FreeBind": b"b",
    b"Transparent": b"b",
    b"Broadcast": b"b",
    b"PassCredentials": b"b",
    b"PassSecurity": b"b",
    b"RemoveOnStop": b"b",
    b"Listen": b"a(ss)",
    b"ListenStream": lambda _, value: (b"Listen", b"a(ss)", [(b"Stream", value)]),
    b"ListenDatagram": lambda _, value: (b"Listen", b"a(ss)", [(b"Datagram", value)]),
    b"ListenSequentialPacket": lambda _, value: (
        b"Listen",
        b"a(ss)",
        [(b"SequentialPacket", value)],
    ),
    b"ListenNetlink": lambda _, value: (b"Listen", b"a(ss)", [(b"Netlink", value)]),
    b"ListenSpecial": lambda _, value: (b"Listen", b"a(ss)", [(b"Special", value)]),
    b"ListenMessageQueue": lambda _, value: (
        b"Listen",
        b"a(ss)",
        [(b"MessageQueue", value)],
    ),
    b"ListenFIFO": lambda _, value: (b"Listen", b"a(ss)", [(b"FIFO", value)]),
    b"ListenUSBFunction": lambda _, value: (
        b"Listen",
        b"a(ss)",
        [(b"USBFunction", value)],
    ),
    b"Symlinks": b"as",
    b"Mark": b"i",
    b"MaxConnections": b"u",
    b"MaxConnectionsPerSource": b"u",
    b"MessageQueueMaxMessages": b"x",
    b"MessageQueueMessageSize": b"x",
    b"TCPCongestion": b"s",
    b"ReusePort": b"b",
    b"SmackLabel": b"s",
    b"SmackLabelIPIn": b"s",
    b"SmackLabelIPOut": b"s",
    b"ControlPID": b"u",
    b"Result": b"s",
    b"NConnections": b"u",
    b"NAccepted": b"u",
    b"NRefused": b"u",
    b"FileDescriptorName": b"s",
    b"SocketProtocol": b"i",
    b"TriggerLimitIntervalUSec": b"t",
    b"TriggerLimitBurst": b"u",
    b"UID": b"u",
    b"GID": b"u",
    b"Id": b"s",
    b"Names": b"as",
    b"Following": b"s",
    b"Upholds": b"as",
    b"UpheldBy": b"as",
    b"OnSuccess": b"as",
    b"OnSuccessOf": b"as",
    b"OnFailureOf": b"as",
    b"PropagatesStopTo": b"as",
    b"StopPropagatedFrom": b"as",
    b"SliceOf": b"as",
    b"AccessSELinuxContext": b"s",
    b"LoadState": b"s",
    b"ActiveState": b"s",
    b"FreezerState": b"s",
    b"SubState": b"s",
    b"FragmentPath": b"s",
    b"DropInPaths": b"as",
    b"UnitFileState": b"s",
    b"UnitFilePreset": b"s",
    b"CanStart": b"b",
    b"CanStop": b"b",
    b"CanReload": b"b",
    b"CanIsolate": b"b",
    b"CanClean": b"as",
    b"CanFreeze": b"b",
    b"OnSuccesJobMode": b"s",
    b"OnSuccessJobMode": b"s",
    b"NeedDaemonReload": b"b",
    b"Markers": b"as",
    b"JobTimeoutUSec": b"t",
    b"JobRunningTimeoutUSec": b"t",
    b"ConditionResult": b"b",
    b"AssertResult": b"b",
    b"Transient": b"b",
    b"Perpetual": b"b",
    b"StartLimitIntervalUSec": b"t",
    b"StartLimitBurst": b"u",
    b"StartLimitAction": b"s",
    b"FailureActionExitStatus": b"i",
    b"SuccessActionExitStatus": b"i",
    b"InvocationID": b"ay",
    b"Refs": b"as",
    b"RequiresOverridable": b"as",
    b"RequisiteOverridable": b"as",
    b"RequiredByOverridable": b"as",
    b"RequisiteOfOverridable": b"as",
    b"StartLimitInterval": b"t",
    b"StartLimitIntervalSec": b"t",
    # Extra
    b"_custom": lambda _, value: value,
}


def signature_array(properties):
    args = [(ord(b"a"), b"(sv)")]
    for prop_name, prop_value in properties.items():
        prop_name = x2char_star(prop_name)
        signature = KNOWN_UNIT_SIGNATURES[prop_name]

        if callable(signature):
            prop_name, signature, prop_value = signature(prop_name, prop_value)

        args += [(ord(b"r"), b"sv"), (ord(b"s"), prop_name)]
        args += [(ord(b"v"), signature)]
        args += apply_signature(signature, [prop_value])
        args += [(-1, None), (-1, None)]
    args += [(-1, None)]

    return args



#!/usr/bin/env python3
#
# Copyright (c) 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in
# the root directory of this source tree.
#

from unittest import TestCase

import pystemd.systemd1.unit_signatures


class TestArraySignature(TestCase):
    def test_sample(self):
        self.assertEqual(
            pystemd.systemd1.unit_signatures.signature_array({"PrivateTmp": True}),
            [
                (ord(b"a"), b"(sv)"),
                (ord(b"r"), b"sv"),
                (ord(b"s"), b"PrivateTmp"),
                (ord(b"v"), b"b"),
                (ord(b"b"), True),
                (-1, None),
                (-1, None),
                (-1, None),
            ],
        )



#!/usr/bin/env python3
#
# Copyright (c) 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in
# the root directory of this source tree.
#


import glob
import subprocess
import sys
from pathlib import Path

from setuptools import setup
from setuptools.extension import Extension

compile_time_env = {"LIBSYSTEMD_VERSION": 500}

try:
    compile_time_env["LIBSYSTEMD_VERSION"] = int(
        subprocess.check_output(["pkg-config", "--modversion", "libsystemd"])
    )
    # Extension has no means to interrogate pkg-config: https://bugs.python.org/issue28207#msg277413
except FileNotFoundError:
    sys.exit(
        '"pkg-config" command could not be found. Please ensure "pkg-config" is installed into your PATH.'
    )
except subprocess.CalledProcessError as e:
    sys.exit(
        "`%s` failed. Please ensure all prerequisite packages from README.md are installed."
        % " ".join(e.cmd)
    )
except ValueError:
    sys.exit("libsystemd version returned by pkg-config is not a plain integer!")

# Use C extensions if respective files are present. Else let Cython modules be
# compiled to C code. The latter is the case when using a clone of the git
# repository, unlike the source distribution which includes both .pyx and .c
# files.
if glob.glob("pystemd/*.c"):
    external_modules = [
        Extension(cext[:-2].replace("/", "."), [cext], libraries=["systemd"])
        for cext in glob.glob("pystemd/*.c")
    ]
else:
    try:
        from Cython.Build import cythonize

        external_modules = cythonize(
            [Extension("*", ["pystemd/*.pyx"], libraries=["systemd"])],
            compile_time_env=compile_time_env,
        )
    except (ImportError, ModuleNotFoundError):
        # If we're just asking for the version, we don't actually need Cython
        if len(sys.argv) == 2 and sys.argv[1] == "--version":
            external_modules = []
            pass
        else:
            raise RuntimeError("Cython not installed.")


setup(
    name="pystemd",
    version="0.13.2",
    author="Alvaro Leiva Geisse",
    author_email="aleivag@gmail.com",
    packages=["pystemd", "pystemd.systemd1", "pystemd.machine1", "pystemd.DBus"],
    ext_modules=external_modules,
    package_data={
        "pystemd": [
            str(p.relative_to("pystemd")) for p in Path("pystemd").glob("**/*.pyi")
        ]
    },
    install_requires=[
        "lxml",
        "psutil",
    ],
    description="A systemd binding for python",
)



# cython: language_level=3
#
# Copyright (c) 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in
# the root directory of this source tree.
#


from libc.stdint cimport uint64_t

cimport pystemd.dbusc as dbusc

from pystemd.utils import x2char_star

LISTEN_FDS_START = dbusc.SD_LISTEN_FDS_START

class PystemdDaemonError(Exception):
  "generic daemon error"


def listen_fds(int unset_environment=0):
  return dbusc.sd_listen_fds(unset_environment)


def notify(int unset_environment, *states, **kwstates):
  """
  interface to systemd sd_notify,

  params:
    unset_environment[int] = 1 if we should remove the notify information,
      otherwise 0.
    states/kwstates = array with states statements that are 'IDENTIFIER=VALUE'.
      e.g: 'READY=1' or keywords states like `ready=1`
      (This will be clear in usage).

  usage:
    * `pystemd.daemon.notify(1, ready=1)`: this will signal `READY=1` to systemd
      notify and will remove all information from the environment. You should not
      try to talk to systemd notify socket again.
    *  `pystemd.daemon.notify(True, "READY=1")`: same as above just passed as a
      string.
    *  `pystemd.daemon.notify(False, ready=1, status='gime gime gime')`: will
      signal systemd that the app is ready and also set the status to
      'gime gime gime'. This command does not clean the notify environment.

    For info on sd_notify, check https://github.com/systemd/systemd/blob/\
master/src/systemd/sd-daemon.h#L173-L232

  """
  pystates = [x2char_star(s) for s in states]
  pystates.extend(
    b'='.join([x2char_star(k.upper()), x2char_star(v, convert_all=True)])
    for k, v in kwstates.items()
  )
  state = b'\n'.join(pystates)
  return dbusc.sd_notify(unset_environment, state)


def booted():
  "Returns True if system was booted with systemd"
  cdef int status = dbusc.sd_booted()
  if status >=0:
    return status > 0

  raise PystemdDaemonError("Could not get systemd booted status")


def watchdog_enabled(int unset_environment=0):
  """
  returns 0 if watchdog is not enabled, and returns the number of usec between
  keep-alive notification messages that the service manager expects.
  """

  cdef:
    int result
    uint64_t usec

  result = dbusc.sd_watchdog_enabled(unset_environment, &usec)
  if result < 0:
    raise PystemdDaemonError('Fail to get watchdog information')

  if result > 0:
    return usec

  return 0



#!/usr/bin/env python3
#
# Copyright (c) 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in
# the root directory of this source tree.
#

import errno
import pathlib
import socket
import tempfile
from unittest import TestCase

from pystemd.dbusexc import DBusConnectionRefusedError, DBusFileNotFoundError
from pystemd.dbuslib import DBusAddress, path_decode, path_encode


class TestEncode(TestCase):
    PREFIX = b"/org/freedesktop/systemd1/unit"
    EXTERNAL_ID = b"s1.service"
    ENCODED_PATH = b"/org/freedesktop/systemd1/unit/s1_2eservice"

    def test_encode(self):
        self.assertEqual(path_encode(self.PREFIX, self.EXTERNAL_ID), self.ENCODED_PATH)

    def test_weird_encode(self):
        self.assertEqual(path_encode(b"/o", self.EXTERNAL_ID), b"/o/s1_2eservice")

    def test_decode(self):
        self.assertEqual(path_decode(self.ENCODED_PATH, self.PREFIX), self.EXTERNAL_ID)

    def test_weird_decode(self):
        self.assertEqual(path_decode(b"/o/s1_2eservice", b"/o"), self.EXTERNAL_ID)


class TestDBusError(TestCase):
    def test_connecting_to_serverless_socket_raises_connection_refused_error(
        self,
    ) -> None:
        with tempfile.TemporaryDirectory() as temporary_directory:
            socket_path = pathlib.Path(temporary_directory) / "test_socket"
            with socket.socket(
                family=socket.AF_UNIX, type=socket.SOCK_STREAM
            ) as server:
                server.bind(bytes(socket_path))

                dbus = DBusAddress(self.dbus_address_from_unix_socket_path(socket_path))
                with self.assertRaises(DBusConnectionRefusedError) as expectation:
                    dbus.open()
                self.assertEqual(expectation.exception.errno, -errno.ECONNREFUSED)

    def test_connecting_to_missing_socket_raises_file_not_found_error(self) -> None:
        with tempfile.TemporaryDirectory() as temporary_directory:
            socket_path = pathlib.Path(temporary_directory) / "test_socket"
            dbus = DBusAddress(self.dbus_address_from_unix_socket_path(socket_path))
            with self.assertRaises(DBusFileNotFoundError) as expectation:
                dbus.open()
            self.assertEqual(expectation.exception.errno, -errno.ENOENT)

    @staticmethod
    def dbus_address_from_unix_socket_path(socket_path: pathlib.Path) -> bytes:
        return b"unix:path=" + bytes(socket_path)



#
# Copyright (c) 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in
# the root directory of this source tree.
#

"""

This library allows you to talk to systemd over dbus from python, without
actually thinking that you are talking to systemd over dbus. This is focus so
you can programmatically, start/stop/restart/kill and verify services status from
systemd point of view, and avoiding `subprocess.Popen(['systemctl', ...`

Usage in interactive mode:

    In [1]: from pystemd.systemd1 import Unit as SDUnit
    In [2]: unit = SDUnit(b'postfix.service')
    In [3]: unit.load()

    # on real code you do In [2,3] in a single line with
    #     with SDUnit(b'postfix.service') as unit:
    #         ... do your code

    In [4]: unit.Unit.ActiveState
    Out[4]: b'active'

    In [5]: unit.Unit.StopWhenUnneeded
    Out[5]: False

    In [6]: unit.Unit.Stop(b'replace') # require privilege account
    Out[6]: b'/org/freedesktop/systemd1/job/6601531'

    In [7]: unit.Unit.ActiveState
    Out[7]: b'inactive'

    In [8]: unit.Unit.Start(b'replace') # require privilege account
    Out[8]: b'/org/freedesktop/systemd1/job/6601532'

    In [9]: unit.Unit.ActiveState
    Out[9]: b'active'

    In [10]: unit.Service.GetProcesses()
    Out[10]:
    [(b'/system.slice/postfix.service',
        1754222,
        b'/usr/libexec/postfix/master -w'),
     (b'/system.slice/postfix.service', 1754224, b'pickup -l -t fifo -u'),
     (b'/system.slice/postfix.service', 1754225, b'qmgr -l -t fifo -u')]

    # Tons of Extra information, examples

    In [11]: unit.Unit.Description
    Out[11]: b'Postfix Mail Transport Agent'

    In [12]: unit.Service.MainPID
    Out[12]: 1754222

    In [13]: unit.Service.PIDFile
    Out[13]: b'/var/spool/postfix/pid/master.pid'


Note a few things:
  1.- This is python3 example.
  2.- We dont really use strings, but bytes.
  3.- Type information is not lost, Out[5] is a boolean, and Out[10] is an
       array of tuples, and each tuple is a (bytes, int, bytes)
  4.- There are a few interfaces to interact, for instance here you saw
      `unit.Unit` and `unit.Service`, There are more interfaces, you can look
      unit.interfaces for a list of them.

      4.1.- Each interface has methods and properties, They can be inspected
        in the commandline, or by inspecting
        `unit.interfaces[<interface>].methods` or
        `unit.interfaces[<interface>].properties`
      4.2.- Properties are read-only.

Extra read (mostly if you want to develop or extend this):

    http://0pointer.net/blog/the-new-sd-bus-api-of-systemd.html : example and
     c code, introduction to busctl.
    https://www.freedesktop.org/wiki/Software/systemd/dbus/ : intro to the
     manager interface.

    some common commandline tools that are useful to know how things are done:

      https://github.com/systemd/systemd/blob/master/src/systemctl/systemctl.c
      https://github.com/systemd/systemd/blob/master/src/busctl/busctl.c
      https://github.com/systemd/systemd/blob/master/src/systemd/sd-bus.h
      https://github.com/systemd/systemd/blob/master/src/libsystemd/sd-bus/sd-bus.c

    similar library but with a different approach (they hardcode all the method
    they want). https://github.com/wiliamsouza/python-systemd . with that said.
    this is build in a way that we can extend the Base Classes to have hardcoded
    methods (example we can have SDUnit.Start that calls SDUnit.Unit.Start)

    some man pages:
        https://www.freedesktop.org/software/systemd/man/index.html#S
"""

from pystemd import DBus, __version__, machine1, systemd1

# handy shortcuts for systemd resources
SDUnit = systemd1.Unit
SDManager = systemd1.Manager
SDMachine = machine1.Machine

__all__ = ["systemd1", "machine1", "__version__", "DBus"]



# cython: language_level=3
#
# Copyright (c) 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in
# the root directory of this source tree.
#

from libc.stdint cimport (
    int16_t,
    int32_t,
    int64_t,
    uint8_t,
    uint16_t,
    uint32_t,
    uint64_t,
)


cdef extern from "sys/uio.h":
    cdef struct iovec:
        void * iov_base
        size_t iov_len


cdef extern from "sys/syscall.h":
    int __NR_setns
    long syscall(long number, ...)


cdef extern from "systemd/sd-journal.h":
    int sd_journal_sendv(iovec *iov, int n);


cdef extern from "systemd/sd-daemon.h":
  int SD_LISTEN_FDS_START
  int sd_listen_fds(int unset_environment)
  int sd_notify(int unset_environment, const char *state)
  int sd_booted()
  int sd_watchdog_enabled(int unset_environment, uint64_t *usec)

cdef extern from "systemd/sd-bus.h":
  ctypedef struct sd_bus:
    pass

  ctypedef struct sd_bus_message:
    pass

  ctypedef struct sd_bus_slot:
    pass

  ctypedef struct sd_bus_error:
    char *name
    char *message
    int _need_free

  ctypedef int (*sd_bus_message_handler_t)(
    sd_bus_message *m,
    void *userdata,
    sd_bus_error *ret_error
  ) except -1

  int sd_bus_open_user(sd_bus **ret)
  int sd_bus_open_system(sd_bus **ret)
  int sd_bus_open_system_remote(sd_bus **ret, const char *host)
  int sd_bus_open_system_machine(sd_bus **ret, const char *machine)

  int sd_bus_default_user(sd_bus **ret)

  int sd_bus_new(sd_bus **ret)
  int sd_bus_set_address(sd_bus *bus, const char *address)
  int sd_bus_set_bus_client(sd_bus *bus, int b)
  int sd_bus_get_address(sd_bus *bus, const char **address)
  int sd_bus_start(sd_bus *ret)

  int sd_bus_call(sd_bus *bus, sd_bus_message *m, uint64_t usec, sd_bus_error *ret_error, sd_bus_message **reply)
  int sd_bus_call_method(
    sd_bus *bus,
    const char *destination,
    const char *path,
    const char *interface,
    const char *member,
    sd_bus_error *ret_error,
    sd_bus_message **reply,
    const char *types, ...)

  int sd_bus_get_property_string(
    sd_bus *bus,
    char *destination,
    char *path,
    char *interface,
    char *member,
    sd_bus_error *ret_error,
    char **ret)

  int sd_bus_get_property(
    sd_bus *bus,
    const char *destination,
    const char *path,
    const char *interface,
    const char *member,
    sd_bus_error *ret_error,
    sd_bus_message **reply,
    const char *type)

  int sd_bus_get_fd(sd_bus *bus)
  int sd_bus_get_unique_name(sd_bus *bus, const char **unique)
  int sd_bus_process(sd_bus *bus, sd_bus_message **r)
  int sd_bus_wait(sd_bus *bus, uint64_t timeout_usec)
  void sd_bus_error_free(sd_bus_error *e)
  sd_bus *sd_bus_unref(sd_bus *bus)
  void sd_bus_close(sd_bus *bus)

  int sd_bus_message_new_method_call(
    sd_bus *bus,
    sd_bus_message **m,
    const char *destination,
    const char *path,
    const char *interface,
    const char *member)

  int sd_bus_message_peek_type(
    sd_bus_message *m,
    char *type,
    const char **contents)

  int sd_bus_message_enter_container(sd_bus_message *m, char type, const char *contents)
  int sd_bus_message_exit_container(sd_bus_message *m)
  int sd_bus_message_open_container(sd_bus_message *m, char type, const char *contents)
  int sd_bus_message_close_container(sd_bus_message *m)
  int sd_bus_message_read(sd_bus_message *m, const char *types, ...)
  int sd_bus_message_read_basic(sd_bus_message *m, char type, void *p)
  int sd_bus_message_append(sd_bus_message *m, const char *types, ...)
  int sd_bus_message_append_array(sd_bus_message *m, char type, const void *ptr, size_t size)

  int sd_bus_message_is_signal(sd_bus_message *m, const char *interface, const char *member)
  int sd_bus_message_is_method_call(sd_bus_message *m, const char *interface, const char *member)
  int sd_bus_message_is_method_error(sd_bus_message *m, const char *name)
  int sd_bus_message_is_empty(sd_bus_message *m)

  int sd_bus_message_get_type(sd_bus_message *m, uint8_t *type)
  int sd_bus_message_get_cookie(sd_bus_message *m, uint64_t *cookie)
  int sd_bus_message_get_reply_cookie(sd_bus_message *m, uint64_t *cookie)
  int sd_bus_message_get_priority(sd_bus_message *m, int64_t *priority)

  int sd_bus_message_get_monotonic_usec(sd_bus_message *m, uint64_t *usec)
  int sd_bus_message_get_realtime_usec(sd_bus_message *m, uint64_t *usec)
  int sd_bus_message_get_seqnum(sd_bus_message *m, uint64_t* seqnum)

  const char *sd_bus_message_get_signature(sd_bus_message *m, int complete)
  const char *sd_bus_message_get_path(sd_bus_message *m)
  const char *sd_bus_message_get_interface(sd_bus_message *m)
  const char *sd_bus_message_get_member(sd_bus_message *m)
  const char *sd_bus_message_get_destination(sd_bus_message *m)
  const char *sd_bus_message_get_sender(sd_bus_message *m)

  sd_bus_message* sd_bus_message_ref(sd_bus_message *m)
  sd_bus_message* sd_bus_message_unref(sd_bus_message *m)

  int sd_bus_path_encode(char* prefix, char* external_id, char **ret_path)
  int sd_bus_path_decode(char* path, char* prefix, char **ret_external_id)

  int sd_bus_set_allow_interactive_authorization(sd_bus *bus, int b);
  int sd_bus_get_allow_interactive_authorization(sd_bus *bus);



cdef extern from "systemd/sd-bus-protocol.h":
  ctypedef enum:
    _SD_BUS_MESSAGE_TYPE_INVALID
    SD_BUS_MESSAGE_METHOD_CALL
    SD_BUS_MESSAGE_METHOD_RETURN
    SD_BUS_MESSAGE_METHOD_ERROR
    SD_BUS_MESSAGE_SIGNAL
    _SD_BUS_MESSAGE_TYPE_MAX

  ctypedef enum:
    _SD_BUS_TYPE_INVALID
    SD_BUS_TYPE_BYTE
    SD_BUS_TYPE_BOOLEAN
    SD_BUS_TYPE_INT16
    SD_BUS_TYPE_UINT16
    SD_BUS_TYPE_INT32
    SD_BUS_TYPE_UINT32
    SD_BUS_TYPE_INT64
    SD_BUS_TYPE_UINT64
    SD_BUS_TYPE_DOUBLE
    SD_BUS_TYPE_STRING
    SD_BUS_TYPE_OBJECT_PATH
    SD_BUS_TYPE_SIGNATURE
    SD_BUS_TYPE_UNIX_FD
    SD_BUS_TYPE_ARRAY
    SD_BUS_TYPE_VARIANT
    SD_BUS_TYPE_STRUCT
    SD_BUS_TYPE_STRUCT_BEGIN
    SD_BUS_TYPE_STRUCT_END
    SD_BUS_TYPE_DICT_ENTRY
    SD_BUS_TYPE_DICT_ENTRY_BEGIN
    SD_BUS_TYPE_DICT_ENTRY_END

cdef union basic_data:
  uint8_t u8;
  uint16_t u16;
  int16_t s16;
  uint32_t u32;
  int32_t s32;
  uint64_t u64;
  int64_t s64;
  double d64;
  const char *string;
  int i;


cdef extern from 'stdbool.h':
  pass

# We may not need this since this file is just to declare where things are.
IF LIBSYSTEMD_VERSION >= 237:
  include "dbusc_v237.pxi"



#!/usr/bin/env python3
#
# Copyright (c) 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in
# the root directory of this source tree.
#

from unittest import TestCase

from pystemd import daemon


class TestNotify(TestCase):
    def test_notify(self):
        notify_status = daemon.notify(
            False, "STATUS=one...", b"STATUS=two....", status="last..."
        )
        self.assertIn(notify_status, {0, 1})


class TestListen(TestCase):
    def test_notify(self):
        self.assertTrue(daemon.listen_fds(False) >= 0)

    def test_listen_fds_start(self):
        self.assertEqual(daemon.LISTEN_FDS_START, 3)
